
!------------------------------------------------------------------------!
!  The Community Multiscale Air Quality (CMAQ) system software is in     !
!  continuous development by various groups and is based on information  !
!  from these groups: Federal Government employees, contractors working  !
!  within a United States Government contract, and non-Federal sources   !
!  including research institutions.  These groups give the Government    !
!  permission to use, prepare derivative works of, and distribute copies !
!  of their work in the CMAQ system to the public and to permit others   !
!  to do so.  The United States Environmental Protection Agency          !
!  therefore grants similar permission to use the CMAQ system software,  !
!  but users are requested to provide copies of derivative works or      !
!  products designed to operate in the CMAQ system to the United States  !
!  Government without restrictions as to use by others.  Software        !
!  that is used with the CMAQ system but distributed under the GNU       !
!  General Public License or the GNU Lesser General Public License is    !
!  subject to their copyright restrictions.                              !
!------------------------------------------------------------------------!

! RCS file, release, date & time of last delta, author, state, [and locker]
! $Header: /project/yoj/arc/CCTM/src/depv/m3dry/m3dry.F,v 1.12 2012/01/19 14:19:43 yoj Exp $

SUBROUTINE m3dry ( iw, abflux, sfc_hono )

!-------------------------------------------------------------------------------
! Name:     Models-3 Dry Deposition
! Purpose:  Computes dry deposition velocities using Rst and Ra, and
!           elements of ADOM DD model.
! Revised:  21 Jan 1998  Original version.  (J. Pleim and A. Bourgeois)
!           18 Sep 2001  Made general for USGS 24-category system.
!                        (T. Otte, J. Pleim, and W. Hutzell)
!           14 Jan 2002  Added temperature dependence to Henry's Law
!                        constants.  Added temperature and pressure
!                        dependence to diffusivity.  Added new dry
!                        deposition species, methanol.  (Y. Wu and T. Otte)
!           18 Jan 2002  Changed the reference wet cuticle resistance.
!                        (J. Pleim)
!           09 Jun 2003  Added logic for modeling snow covered surfaces.
!                        Changed the reactivities for SO2, HNO3 and NH3.
!                        Changed pH values to have an east-west variation.
!                        Using the Henry's law constant function from CMAQ in
!                        place of local code.  Also changed the code for
!                        deposition to water to use a pH of 8.1 and the
!                        temperature of water in calculating the Henry's law
!                        constant.  Adjusted values of RSNOW0 = 1000 and
!                        A(NH3) = 20.  Added new dry deposition species: N2O5,
!                        NO3, Generic_aldehyde.  Corrected diffusivities of
!                        chemicals and water and viscosity of air to all be at
!                        the same temperature (273.15K).  Temperature and
!                        pressure adjustments to the values are not needed
!                        because the diffusivities and viscosity are always used
!                        as ratios, so the temperature-pressure dependence was
!                        removed.  Removed dry deposition species, ATRA and
!                        ATRAP, from output.  (D. Schwede, J. Pleim, and
!                        T. Otte)
!           28 Feb 2005  Added optional dry deposition species for chlorine
!                        and mercury.  (G. Sarwar, R. Bullock, and T. Otte)
!           02 Feb 2006  Added mesophyll resistance to dry deposition velocity
!                        calculation, and defined non-zero value for mercury.
!                        (D. Schwede, J. Pleim, and R. Bullock)
!           01 Aug 2007  Added a non-zero mesophyll resistance for NO, NO2, and
!                        CO.  Restored wet cuticle resistance for O3 based on
!                        field study measurements.  Added wet ground resistance.
!                        Changed ground resistance to include partitioning of
!                        wet and dry ground.  Updated pH of rain water for
!                        eastern United States and outside of North America.
!                        Changed reactivity for PAN.  Removed dry deposition
!                        velocity calculations for obsolete chlorine species
!                        ICL1 and ICL2.  Corrected error in the calculation of
!                        surface resistance over water where (Sc/Pr)**(2/3) had
!                        been inadvertently omitted from the numerator.
!                        Surface resistance over water is now a function of
!                        species.  Surface resistance over water now uses wet
!                        bulb temperature rather than ground (water) temperature
!                        in the calculation of the effective Henry's law
!                        constant, and the algorithm has been updated.  Changed
!                        (Sc/Pr)**(2/3) over water to a species-dependent,
!                        meteorologically dependent variable.  Effective Henry's
!                        law constant over land now uses 2-m temperature rather
!                        than layer 1 temperature. Changed ES
!                        into ES_AIR and ES_GRND, and changed QSS into QSS_AIR
!                        and QSS_GRND to clarify usage.  (J. Pleim, E. Cooter,
!                        J. Bash, T. Otte, and G. Sarwar)
!           07 Dec 2007  Add into CMAQ for in-line deposition velocities.
!                        (W. Hutzell, J. Young and T. Otte)
!           07 Jan 2008  Changed the value of d3, the scaling parameter used to
!                        estimate the friction velocity in surface waters from
!                        the atmospheric friction velocity to a value following
!                        Slinn et al. (1978) and Fairall et al. (2007).
!                        (J. Bash)
!           01 Feb 2008  Added bidirectional NH3 flux calculations. (J. Pleim and
!                        J. Young)
!           20 Mar 2008  Added a trap for undefined dry deposition velocities
!                        (e.g., NaN's).  (T. Otte)
!           21 Mar 2008  Added heterogeneous reaction for HONO. It affects HONO, NO2
!                        and HNO3 (G. Sarwar)
!           30 Apr 2008  Added five air toxic species to output.  (W. Hutzell
!                        and T. Otte)
!           August 2008  Applied a minimum value to ustg (0.001 m/s) to prevent 
!                        negative rbg values in the bidi calculation (J. Pleim) 
!           05 Oct 2009  Added condition that vegetation fraction must be
!                        greater than zero to be considered a land point.  This
!                        works around intermittent inconsistencies in surface
!                        fields in some WRF data sets.  (T. Otte)
!           Dec 2009     Revised bidirectional NH3 flux calculations to use soil 
!                        Gamma values read from gridded file.  Bidi flux calcs
!                        based on comparisons with Lillington, NC corn data 2007 (J. Pleim)
!           30 Mar 2010  Modified to output the NH3 bidi stomtal, cuticle and soil component 
!                        fluxes and chaged NH3 bidi variables used in estimating the compensation
!                        point double. (J. Bash)
!           16 Feb 2011 S.Roselle: replaced I/O API include files with UTILIO_DEFN
!           20 May 2011  D.Schwede: add MOSAIC processing
!           14 Jul 2011  Replaced dw25 calculation with Hayduk and Laudie method.
!                        LeBas molar volumes are from the Schroeder additive method
!                        with the exception of HGIIGAS (modeled as HgCl2) which was 
!                        obtained using the Tyn and Calus method. Also, ICL1 and ICL2
!                        were removed. (D. Schwede)
!           27 Jul 2011 J.Bash: Parmaterized the mesophyll resistance as a function
!                        of solubility following Wesely 1989 Atmos Environ.
!           15 Aug 2011  Modified HONO calculation so that deposition velocity for NO2
!                        that is output in DEPV file does not include the loss due to
!                        the heterogeneous reaction. This additional loss is now
!                        accounted for in vdiff.F (D. Schwede and G. Sarwar)
!           29 Aug 2011 Added NH3 bidirectional flux variables and modules and integrated
!                       NH3 bidi algorithms with MOSAIC algorithms. NH3 bidi routines now
!                       read in foratted EPIC output and maintain a soil NH4 budget and 
!                       fluxes are calculated for individual land cover types. 
!                       (J. Bash and D. Schwede)
!           22 Sep 2011 -- incorporated twoway model implementation
!                       -- removed non-use dluse array
!                          (David Wong)
!           26 Sep 2011 -- made the number of actual and dummy arguments the same in
!                          calling subroutine Init_ABFlux
!                          (David Wong)
!-------------------------------------------------------------------------------

  use depvvars
  use depv_defn,   only: depvel_gas_land, depvel_gas_sea, ie_hono, ic_no2
  use utilio_defn
  use const_data
  use cgrid_defn,  only: cgrid, vdemis_gc
  use mem_ijtabs,  only: itab_w
  use mem_leaf,    only: land, itab_wl
  use mem_sea,     only: sea,  itab_ws
  use mem_grid,    only: dzt, volti
  use mem_basic,   only: theta, rho
  use mem_turb,    only: pblh
  use leaf_coms,   only: nzg, slcpd, slmsts_ch, slmsts_vg
  use hlconst_mod, only: hlconst, hlconst_spcs_init, is_effect_spc
  use micro_coms,  only: ccnparm
  use mem_micro,   only: cldnum
  use therm_lib,   only: rhovsil, qtk, qtk_sea, qwtk

  implicit none

! Arguments:

  integer,     intent( in )  :: iw
  logical,     intent( in )  :: abflux
  logical,     intent( in )  :: sfc_hono
      
! Local Variables:

  REAL,            PARAMETER :: a0         = 8.0     ! [dim'less]
  REAL,            PARAMETER :: d3         = 1.38564e-2 ! [dim'less]
                                               ! k*sqrt(rhoair/rhowater) from Slinn 78
  real                       :: d_ice      ! dep. vel. over ice
  REAL                       :: dw
  REAL,            PARAMETER :: dwat       = 0.2178  ! [cm^2/s] at 273.15K
  LOGICAL                    :: effective  ! true=compute effective Henry's Law const

  LOGICAL, SAVE              :: first_call = .TRUE.
  REAL                       :: heff                 ! effective Henry's Law constant
  REAL                       :: heff_ap              ! Henry's Law constant for leaf apoplast M/atm
  REAL                       :: hplus
  REAL,            PARAMETER :: hplus_ap    = 1.0e-6     ! pH=6.0 leaf apoplast solution Ph (Massad et al 2008)
  REAL,            PARAMETER :: hplus_dirty = 1.0e-5     ! pH=5.0
  REAL,            PARAMETER :: hplus_clean = 3.16228e-6 ! 10.0**(-5.5)
  REAL,            PARAMETER :: hplus_h2o   = 7.94328e-9 ! 10.0**(-8.1)
  REAL,            PARAMETER :: hplus_def   = 1.0e-5     ! pH=5.0
  real                       :: hveg
  LOGICAL                    :: ifurban
  LOGICAL                    :: isocean
  integer                    :: iwl, jwl
  integer                    :: iws, jws
  REAL,            PARAMETER :: kvis       = 0.132   ! [cm^2 / s] at 273.15K
  REAL                       :: kviswi      ! 1 / kinematic viscosity of water [s/cm^2]
  INTEGER                    :: kw
  INTEGER                    :: l
  REAL                       :: lai, laiv          ! leaf area index
  REAL                       :: tai, taiv          ! total area index
  INTEGER                    :: n
  CHARACTER( 16 ), PARAMETER :: pname      = 'M3DRY'
  REAL,            PARAMETER :: pr         = 0.709   ! [dim'less]
  real                       :: pvd
  real                       :: rai                  ! cell aerodynamic resistance
  real                       :: raicei               ! aerodynamic resistance over ice
  real                       :: rawi                 ! aerodynamic resistance over water
  REAL                       :: rbci
  REAL                       :: rbsulfi
  REAL                       :: rci
  REAL,            PARAMETER :: rcut0      = 3000.0   ! [s/m]
  REAL,            PARAMETER :: rcw0       = 125000.0 ! adapted from Slinn 78
  REAL,            PARAMETER :: rcw0i      = 1.0 / rcw0
  
  REAL,            PARAMETER :: rg0        = 1000.0  ! [s/m]
  REAL                       :: rgndi
  REAL                       :: rgwi                 ! resist for water-covered sfc

  real,            parameter :: cwetO3 = 1.0 / 1250.0 ! O3 wet cuticle conductance, Altimir et al 2006

  REAL,            PARAMETER :: rgwet0     = 25000.0      ! [s/m]
  REAL,            PARAMETER :: rgwet0i    = 1.0 / rgwet0 ! m/s]

  REAL                       :: rh_air               ! rel humidity (air)
  REAL                       :: rinc                 ! resitanced from ground through canopy
  REAL                       :: rsi                  ! stomatal resistance [s/m]
  REAL,            PARAMETER :: rsndiff    = 10.0    ! snow diffusivity fac
  REAL                       :: rsnowi
  REAL,            PARAMETER :: rsnow0     = 1000.0
  REAL                       :: rstomi
  REAL                       :: rsurfi
  REAL                       :: rweti                ! wet sfc resist (cuticle or grnd)
  REAL                       :: rwetsfci
  REAL                       :: scw_pr_23            ! (scw/pr)**2/3
  real                       :: sst
  REAL,            PARAMETER :: svp2       = 17.67   ! from MM5 and WRF
  REAL,            PARAMETER :: svp3       = 29.65   ! from MM5 and WRF
  REAL,            PARAMETER :: rt25inK    = 1.0/(stdtemp + 25.0) ! 298.15K = 25C
  REAL                       :: tempcr               ! cell canopy temp
  REAL                       :: tempgr               ! cell ground temp
  REAL                       :: tice
  REAL                       :: tw
  REAL,            PARAMETER :: onethird   = 1.0 / 3.0
  REAL,            PARAMETER :: twothird   = 2.0 / 3.0
  REAL                       :: ustar                ! cell friction velocity
  REAL                       :: ustari               ! friction velocity over ice
  REAL                       :: ustarw               ! friction velocity over water
  REAL                       :: vegfr                ! cell veg coverage fraction
  real                       :: wrcr                 ! vegetation water (kg/m^2)
  REAL                       :: wrmax
  REAL                       :: wstar
  REAL                       :: wtv0
  REAL                       :: xm                   ! liquid water mass frac
  REAL                       :: xt                   ! liquid water mass frac
  real                       :: rh_func              ! RH function for the development of a water
                                                         !      film on leaf cuticles
  real                       :: slmsts
  real                       :: slmst1

  real :: tempvg
  real :: deltag, liqfrg
  real :: deltav, liqfrv
  real :: fact
  real :: rgliqi
  real :: rdryi
  real :: rcuti
  real :: rgdi
  real :: rgndci
  real :: rmesoi

  INTEGER, PARAMETER         :: n_spc_m3dry = ltotg       ! from DEPVVARS module

  REAL, save                 :: ar       ( ltotg )        ! reactivity relative to HNO3
  REAL, save                 :: dif0     ( ltotg )        ! molecular diffusivity [cm2/s]
  REAL, save                 :: ddif0    ( ltotg )        ! molecular diffusivity / dwat
  REAL, save                 :: lebas    ( ltotg )        ! Le Bas molar volume [cm3/mol ]
  REAL, save                 :: meso     ( ltotg )        ! Exception for species that 
                                                          ! react with cell walls.
                                                          ! Wesely 1989 eq 6.
  REAL, SAVE                 :: scc_pr_23( ltotg )        ! 0.2(PR/SCC)**2/3, fn of DIF0
  real, save                 :: csnow    ( ltotg )
  real, save                 :: cgnd     ( ltotg )
  real, save                 :: ccut     ( ltotg )
  REAL, save                 :: dw25     ( ltotg )     ! diffusivity of water at 25C


  CHARACTER( 16 ), save      :: subname  ( ltotg )        ! for subroutine HLCONST
  integer, save              :: hlspc    ( ltotg )        ! species index in hlconst
  logical, save              :: hleff    ( ltotg )

!-------------------------------------------------------------------------------
! For ozone exchange over the ocean
  REAL                       :: pchang      ! p value used in equation (12) in Chang et al( 2004)
  REAL                       :: kwchang     ! kw value used in Chang et al (2004)
  REAL                       :: ciodide     ! iodide concentration (umol/L)
  REAL                       :: qiodide     ! q in Chang et al (2004)

!-------------------------------------------------------------------------------
! For heterogenous hono on surfaces
  REAL                       :: kno2        ! first order rate constant for the heterogenous reaction [1/s]  
  REAL                       :: surf_bldg   ! Surface area of bldgs to volume of air [-]
  REAL                       :: surf_leaf   ! Surface area of leaves to volume of air [-]
  REAL                       :: zf          ! layer height [m]      
      
!-------------------------------------------------------------------------------
! For Ammonia bi-directional flux                      
! REAL                       :: sltyp     ! soil type
! REAL                       :: wg        ! soil moisture in top 1 cm (vol frc)
! REAL                       :: w2        ! soil moisture in top 1 m (vol frc)
! REAL                       :: cnh3      ! Layer 1 NH3 concentration from CGRID [ppm]
! REAL                       :: tpvd      ! temp pvd variable
! REAL                       :: f_stom    ! stomatal flux component
! REAL                       :: f_cut     ! cuticular flux component
! REAL                       :: f_soil    ! soil flux component
! REAL                       :: f_emis    ! Emission flux component
! REAL                       :: f_dep     ! Deposition flux component
! REAL                       :: f_ag      ! flux from agriculture
! REAL                       :: f_nat     ! flux from natural systems
! REAL                       :: f_wat     ! direct flux to surface water
! REAL                       :: lnh3      ! loss part of ammonia flux
                                          ! Note that lnh3 is stored in DEPVEL_GAS(nh3) for use in vdiff

!-------------------------------------------------------------------------------
!-- Chemical-Dependent Parameters (Original Source: Modified ADOM - Padro)
!                                                        at 298.15 K
!     Species   Dif0(cm2/s) Alphastar   Reactivity     -DKHOR [K]  KH0 [M/atm]
!     _______   ___________ _________   __________     _________  __________
!  1   SO2      0.1089 ~    1000.       8.00            3100.0 ~     1.2e00 ~
!  2   SULF      --       --           --                --           --
!  3   NO2      0.1361 ~    1.00      8.00 2.00*        2500.0 ~     1.2e-2 ~
!  4   NO       0.1802 ~    1         2.0+              1500.0 ~     1.9e-3 ~
!  5   O3       0.1444 ~    10.00       15.0  8@        2700.0 ~     1.2e-2 ~
!  6   HNO3     0.1628 ~    1E9       18.0 800* 8000**  8700.0 ~     2.6e+6 ~
!  7   H2O2     0.2402 ~    1.00      12.0 30*          6600.0 ~     9.7e+4 ~
!  8   ACET ALD 0.1525 ~    1         10.+              5700.0 ~     1.3e+1 ~
!  9   HCHO     0.1877 ~    1.00      10.+              6400.0 ~     7.0e+3 ~
! 10   OP       0.1525 ~    1         10.0+             5200.0 ~     3.1e+2 ~
! 11   PAA      0.1220 ~    1         20+               5300.0 ~     8.4e+2 ~
! 12   ORA      0.1525 ~    1         20+               5700.0 ~     3.7e+3 ~
! 13   NH3      0.1978 ~    1E5       10.0              4100.0 ~     5.8e+1 ~
! 14   PAN      0.0938 ~    1         16.0~~            5900.0 ~     2.9e00 ~
! 15   HONO     0.1525 ~    1         20+               4800.0 ~     4.9e+1 ~
! 16   CO       0.1807 ~    1         5.0               1600.0 ~     9.5e-4 ~
! 17   METHANOL 0.1363 ~    1.0 ~       2.0 ~           5200.0 ~     2.2e+2 ~
! --   CO2      0.1381 ~                                2400.0 ~     3.4e-2 ~
! 18   N2O5     0.0808 ^^             5000.0**
! 19   NO3      0.1153 ^^             5000.0**
! 20   GEN ALD  0.1525 ##   1.0       10.0##
! 21   CL2      0.1080 %              10.0 %
! 22   HOCL     0.1300 %              10.0 %
! 23   HCL      0.1510 %              8000.0 %
! 24   FMCL     0.1094 %              10.0 %
! 27   HG       0.1194 $              0.1 $
! 28   HGIIGAS  0.0976 $              8000.0 $
! 50   HEXAMETHYLENE DIISOCYANATE       10.0 <>
! 51   HYDRAZINE                        20.0 <>
! 52   MALEIC ANHYDRIDE                 10.0 <>
! 53   TOLUENE DIISOCYANATE             10.0 <>
! 54   TRIETHYLAMINE                    10.0 <>
!
!---------Notes
!  * Updates based on literature review 7/96 JEP
!  # Diff and H based on Wesely (1988) same as RADM
!  + Estimated by JEP 2/97
!  @ Updated by JEP 9/01
!  ~ Added by YW 1/02.  Dif0 based on Massman (1998).  Henry's Law constant
!    is defined here as: h=cg/ca, where cg is the concentration of a species
!    in gas-phase, and ca is its aqueous-phase concentration.  The smaller h,
!    the larger solubility.  Henry's Law constant in another definition (KH):
!    KH = ca/pg [M/atm], KH = KH0 * exp(-DKH/R(1/T-1/T0)), where KH0 and -DKH
!    values are from Rolf Sander (1999).  h=1/(KH*R*T).
! ** Update by DBS based on estimates by JEP 1/03
! ^^ From Bill Massman, personal communication 4/03
! ## Diffusivity calculated by SPARC, reactivity = other aldehydes
! ++ Dif0 in Massman is diffusivity at temperature 0C and 1 atm (101.325kPa), so
!    chemicals that were not in Massman's paper need to be adjusted.  We assume
!    JEP's original values were for 25C and 1 atm.
!  % Added by G. Sarwar (10/04)
!  $ Added by R. Bullock (02/05) HG diffusivity is from Massman (1999).
!    HGIIGAS diffusivity calculated from the HG value and a mol. wt. scaling
!    factor of MW**(-2/3) from EPA/600/3-87/015. ORD, Athens, GA.  HGIIGAS
!    mol.wt. used is that of HgCl2.  Reactivity of HG is 1/20th of NO and NO2
!    values based on general atmospheric lifetimes of each species.  Reactivity
!    of HGIIGAS is based on HNO3 surrogate.
! @@ Mesophyll resistances for NO, NO2, and CO added by J. Pleim (07/07) based
!    on values in Pleim, Venkatram, and Yamartino, 1984:  ADOM/TADAP Model
!    Development Program, Volume 4, The Dry Deposition Module.  ERT, Inc.,
!    Concord, MA (peer reviewed).
! ~~ Reactivity for PAN changed from 4.0 to 16.0 by J. Pleim (07/07) based on
!    comparisons with Turnipseed et al., JGR, 2006.
! %% Species ICL1 and ICL2 are removed, not used in CB05.  G. Sarwar (07/07)
! <> Hazardous Air Pollutants that are believed to undergo significant dry
!    deposition. Hydrazine and triethylamine reactivities are based on analogies
!    to NH3. Maleic anhydride reactivity is assumed similar to aldehydes.
!    Toluene diisocyanate and hexamethylene diisocyanate reactivities are
!    assumed to be similar to SO2. Diffusivities are calculated with standard
!    formulas.  W. Hutzell (04/08)
! %% G. Sarwar: added data for iodine and bromine species (03/2016)
! %% B. Hutzell: added dry deposition data for methane, acrylic acid, methyl chloride,
!    and acetonitrile (09/2016)   

!-------------------------------------------------------------------------------

  DATA subname(  1), dif0(  1), ar(  1), meso(  1), lebas(  1) / 'SO2             ', 0.1089,   10.0,      0.0,  35.0/
  DATA subname(  2), dif0(  2), ar(  2), meso(  2), lebas(  2) / 'H2SO4           ', 0.1091, 8000.0,      0.0,  49.0/
  DATA subname(  3), dif0(  3), ar(  3), meso(  3), lebas(  3) / 'NO2             ', 0.1361,    2.0,      0.1,  21.0/
  DATA subname(  4), dif0(  4), ar(  4), meso(  4), lebas(  4) / 'NO              ', 0.1802,    2.0,      0.0,  14.0/
  DATA subname(  5), dif0(  5), ar(  5), meso(  5), lebas(  5) / 'O3              ', 0.1444,   12.0,      1.0,  21.0/
  DATA subname(  6), dif0(  6), ar(  6), meso(  6), lebas(  6) / 'HNO3            ', 0.1067, 8000.0,      0.0,  35.0/
  DATA subname(  7), dif0(  7), ar(  7), meso(  7), lebas(  7) / 'H2O2            ', 0.1300,34000.0,      1.0,  28.0/   !ar=34,000 such that r_cut=0.7 s/m as in Nguyen et al. 2015
  DATA subname(  8), dif0(  8), ar(  8), meso(  8), lebas(  8) / 'ACETALDEHYDE    ', 0.1111,   10.0,      0.0,  56.0/
  DATA subname(  9), dif0(  9), ar(  9), meso(  9), lebas(  9) / 'FORMALDEHYDE    ', 0.1554,   10.0,      0.0,  35.0/
  DATA subname( 10), dif0( 10), ar( 10), meso( 10), lebas( 10) / 'METHYLHYDROPEROX', 0.1179,   10.0,      0.3,  49.0/   !meso change from 0.1 to 0.3, Wolfe and Thornton 2011 ACP per J. Bash
  DATA subname( 11), dif0( 11), ar( 11), meso( 11), lebas( 11) / 'PEROXYACETIC_ACI', 0.0868,   20.0,      0.1,  70.0/
  DATA subname( 12), dif0( 12), ar( 12), meso( 12), lebas( 12) / 'ACETIC_ACID     ', 0.0944,   20.0,      0.0,  63.0/
  DATA subname( 13), dif0( 13), ar( 13), meso( 13), lebas( 13) / 'NH3             ', 0.1978,   20.0,      0.0,  28.0/
  DATA subname( 14), dif0( 14), ar( 14), meso( 14), lebas( 14) / 'PAN             ', 0.0687,   16.0,      0.1,  91.0/
  DATA subname( 15), dif0( 15), ar( 15), meso( 15), lebas( 15) / 'HNO2            ', 0.1349,   20.0,      0.1,  28.0/
  DATA subname( 16), dif0( 16), ar( 16), meso( 16), lebas( 16) / 'CO              ', 0.1807,    5.0,      0.0,  14.0/
  DATA subname( 17), dif0( 17), ar( 17), meso( 17), lebas( 17) / 'METHANOL        ', 0.1329,    2.0,      0.0,  42.0/
  DATA subname( 18), dif0( 18), ar( 18), meso( 18), lebas( 18) / 'N2O5            ', 0.0808, 5000.0,      0.0,  49.0/
  DATA subname( 19), dif0( 19), ar( 19), meso( 19), lebas( 19) / 'NO3             ', 0.1153, 5000.0,      0.0,  28.0/
  DATA subname( 20), dif0( 20), ar( 20), meso( 20), lebas( 20) / 'GENERIC_ALDEHYDE', 0.0916,   10.0,      0.0,  56.0/
  DATA subname( 21), dif0( 21), ar( 21), meso( 21), lebas( 21) / 'CL2             ', 0.1080,   10.0,      0.0,  49.0/
  DATA subname( 22), dif0( 22), ar( 22), meso( 22), lebas( 22) / 'HOCL            ', 0.1300,   10.0,      0.0,  38.5/
  DATA subname( 23), dif0( 23), ar( 23), meso( 23), lebas( 23) / 'HCL             ', 0.1510, 8000.0,      0.0,  31.5/
  DATA subname( 24), dif0( 24), ar( 24), meso( 24), lebas( 24) / 'FMCL            ', 0.1094,   10.0,      0.0,  45.5/
  DATA subname( 25), dif0( 25), ar( 25), meso( 25), lebas( 25) / 'HG              ', 0.1194,    0.1,      0.0,  14.8/   ! lebas not used
  DATA subname( 26), dif0( 26), ar( 26), meso( 26), lebas( 26) / 'HGIIGAS         ', 0.0976, 8000.0,      0.0,  95.0/   ! estimation from back calculating to get dw25 = 1.04e-5 (Garland et al, 1965)
  DATA subname( 27), dif0( 27), ar( 27), meso( 27), lebas( 27) / 'TECDD_2378      ', 0.0525,    2.0,      0.0, 217.0/
  DATA subname( 28), dif0( 28), ar( 28), meso( 28), lebas( 28) / 'PECDD_12378     ', 0.0508,    2.0,      0.0, 234.5/
  DATA subname( 29), dif0( 29), ar( 29), meso( 29), lebas( 29) / 'HXCDD_123478    ', 0.0494,    2.0,      0.0, 252.0/
  DATA subname( 30), dif0( 30), ar( 30), meso( 30), lebas( 30) / 'HXCDD_123678    ', 0.0494,    2.0,      0.0, 252.0/
  DATA subname( 31), dif0( 31), ar( 31), meso( 31), lebas( 31) / 'HXCDD_123478    ', 0.0494,    2.0,      0.0, 252.0/
  DATA subname( 32), dif0( 32), ar( 32), meso( 32), lebas( 32) / 'HPCDD_1234678   ', 0.0480,    2.0,      0.0, 269.5/
  DATA subname( 33), dif0( 33), ar( 33), meso( 33), lebas( 33) / 'OTCDD           ', 0.0474,    2.0,      0.0, 287.0/
  DATA subname( 34), dif0( 34), ar( 34), meso( 34), lebas( 34) / 'TECDF_2378      ', 0.0534,    2.0,      0.0, 210.0/
  DATA subname( 35), dif0( 35), ar( 35), meso( 35), lebas( 35) / 'PECDF_12378     ', 0.0517,    2.0,      0.0, 227.5/
  DATA subname( 36), dif0( 36), ar( 36), meso( 36), lebas( 36) / 'PECDF_23478     ', 0.0517,    2.0,      0.0, 227.5/
  DATA subname( 37), dif0( 37), ar( 37), meso( 37), lebas( 37) / 'HXCDF_123478    ', 0.0512,    2.0,      0.0, 245.0/
  DATA subname( 38), dif0( 38), ar( 38), meso( 38), lebas( 38) / 'HXCDF_123678    ', 0.0512,    2.0,      0.0, 245.0/
  DATA subname( 39), dif0( 39), ar( 39), meso( 39), lebas( 39) / 'HXCDF_234678    ', 0.0512,    2.0,      0.0, 245.0/
  DATA subname( 40), dif0( 40), ar( 40), meso( 40), lebas( 40) / 'HXCDF_123789    ', 0.0512,    2.0,      0.0, 245.0/
  DATA subname( 41), dif0( 41), ar( 41), meso( 41), lebas( 41) / 'HPCDF_1234678   ', 0.0487,    2.0,      0.0, 262.5/
  DATA subname( 42), dif0( 42), ar( 42), meso( 42), lebas( 42) / 'HPCDF_1234789   ', 0.0487,    2.0,      0.0, 262.5/
  DATA subname( 43), dif0( 43), ar( 43), meso( 43), lebas( 43) / 'OTCDF           ', 0.0474,    2.0,      0.0, 280.0/
  DATA subname( 44), dif0( 44), ar( 44), meso( 44), lebas( 44) / 'NAPHTHALENE     ', 0.0778,    4.0,      0.0, 119.0/
  DATA subname( 45), dif0( 45), ar( 45), meso( 45), lebas( 45) / '1NITRONAPHTHALEN', 0.0692,    4.0,      0.0, 133.0/
  DATA subname( 46), dif0( 46), ar( 46), meso( 46), lebas( 46) / '2NITRONAPHTHALEN', 0.0692,    4.0,      0.0, 133.0/
  DATA subname( 47), dif0( 47), ar( 47), meso( 47), lebas( 47) / '14NAPHTHOQUINONE', 0.0780,    4.0,      0.0, 119.0/
  DATA subname( 48), dif0( 48), ar( 48), meso( 48), lebas( 48) / 'HEXAMETHYLE_DIIS', 0.0380,   10.0,      0.0, 196.0/
  DATA subname( 49), dif0( 49), ar( 49), meso( 49), lebas( 49) / 'HYDRAZINE       ', 0.4164,   20.0,      0.0,  42.0/
  DATA subname( 50), dif0( 50), ar( 50), meso( 50), lebas( 50) / 'MALEIC_ANHYDRIDE', 0.0950,   10.0,      0.0,  70.0/
  DATA subname( 51), dif0( 51), ar( 51), meso( 51), lebas( 51) / '24-TOLUENE_DIIS ', 0.0610,   10.0,      0.0, 154.0/
  DATA subname( 52), dif0( 52), ar( 52), meso( 52), lebas( 52) / 'TRIETHYLAMINE   ', 0.0881,   20.0,      0.0, 154.0/
  DATA subname( 53), dif0( 53), ar( 53), meso( 53), lebas( 53) / 'ORG_NTR         ', 0.0607,   16.0,      0.0, 160.0/  ! assumes 58.2% C5H11O4N and 41.8% C5H11O3N
  DATA subname( 54), dif0( 54), ar( 54), meso( 54), lebas( 54) / 'HYDROXY_NITRATES', 0.0609,   16.0,      0.0, 156.1/
  DATA subname( 55), dif0( 55), ar( 55), meso( 55), lebas( 55) / 'MPAN            ', 0.0580,   16.0,      0.1, 133.0/
  DATA subname( 56), dif0( 56), ar( 56), meso( 56), lebas( 56) / 'PPN             ', 0.0631,   16.0,      0.1, 118.2/
  DATA subname( 57), dif0( 57), ar( 57), meso( 57), lebas( 57) / 'MVK             ', 0.0810,    8.0,      1.0,  88.8/
  DATA subname( 58), dif0( 58), ar( 58), meso( 58), lebas( 58) / 'DINTR           ', 0.0617,   16.0,      0.1, 169.8/
  DATA subname( 59), dif0( 59), ar( 59), meso( 59), lebas( 59) / 'NTR_ALK         ', 0.0688,   16.0,      0.1, 133.0/
  DATA subname( 60), dif0( 60), ar( 60), meso( 60), lebas( 60) / 'NTR_OH          ', 0.0665,   16.0,      0.1, 140.4/
  DATA subname( 61), dif0( 61), ar( 61), meso( 61), lebas( 61) / 'HYDROXY_NITRATES', 0.0646,   16.0,      0.0, 147.8/
  DATA subname( 62), dif0( 62), ar( 62), meso( 62), lebas( 62) / 'PROPNN          ', 0.0677,   16.0,      0.0, 133.0/
  DATA subname( 63), dif0( 63), ar( 63), meso( 63), lebas( 63) / 'NITRYL_CHLORIDE ', 0.0888,    8.0,      0.0,  45.5/  ! dif0 estimated following Erickson III et al., JGR, 104, D7, 8347-8372, 1999
  DATA subname( 64), dif0( 64), ar( 64), meso( 64), lebas( 64) / 'ISOPNN          ',0.0457,    8.0,      0.0,  206.8/
  DATA subname( 65), dif0( 65), ar( 65), meso( 65), lebas( 65) / 'MTNO3           ',0.0453,    8.0,      0.0,  251.2/
  DATA subname( 66), dif0( 66), ar( 66), meso( 66), lebas( 66) / 'IEPOX           ',0.0579,    8.0,      0.0,  110.8/
  DATA subname( 67), dif0( 67), ar( 67), meso( 67), lebas( 67) / 'HACET           ',0.1060,    8.0,      0.0,   72.6/  ! dif0 from Nguyen 2015 PNAS
  DATA subname( 68), dif0( 68), ar( 68), meso( 68), lebas( 68) / 'SVALK1          ',0.0514,   20.0,      0.0,  280.5/
  DATA subname( 69), dif0( 69), ar( 69), meso( 69), lebas( 69) / 'SVALK2          ',0.0546,   20.0,      0.0,  275.6/
  DATA subname( 70), dif0( 70), ar( 70), meso( 70), lebas( 70) / 'SVBNZ1          ',0.0642,   20.0,      0.0,  134.1/
  DATA subname( 71), dif0( 71), ar( 71), meso( 71), lebas( 71) / 'SVBNZ2          ',0.0726,   20.0,      0.0,  127.5/
  DATA subname( 72), dif0( 72), ar( 72), meso( 72), lebas( 72) / 'SVISO1          ',0.0733,   20.0,      0.0,  126.3/
  DATA subname( 73), dif0( 73), ar( 73), meso( 73), lebas( 73) / 'SVISO2          ',0.0729,   20.0,      0.0,  123.8/
  DATA subname( 74), dif0( 74), ar( 74), meso( 74), lebas( 74) / 'SVPAH1          ',0.0564,   20.0,      0.0,  235.7/
  DATA subname( 75), dif0( 75), ar( 75), meso( 75), lebas( 75) / 'SVPAH2          ',0.0599,   20.0,      0.0,  231.5/
  DATA subname( 76), dif0( 76), ar( 76), meso( 76), lebas( 76) / 'SVSQT           ',0.0451,   20.0,      0.0,  346.5/
  DATA subname( 77), dif0( 77), ar( 77), meso( 77), lebas( 77) / 'SVTOL1          ',0.0637,   20.0,      0.0,  153.7/
  DATA subname( 78), dif0( 78), ar( 78), meso( 78), lebas( 78) / 'SVTOL2          ',0.0607,   20.0,      0.0,  194.1/
  DATA subname( 79), dif0( 79), ar( 79), meso( 79), lebas( 79) / 'SVTRP1          ',0.0603,   20.0,      0.0,  194.9/
  DATA subname( 80), dif0( 80), ar( 80), meso( 80), lebas( 80) / 'SVTRP2          ',0.0559,   20.0,      0.0,  218.8/
  DATA subname( 81), dif0( 81), ar( 81), meso( 81), lebas( 81) / 'SVXYL1          ',0.0610,   20.0,      0.0,  154.6/
  DATA subname( 82), dif0( 82), ar( 82), meso( 82), lebas( 82) / 'SVXYL2          ',0.0585,   20.0,      0.0,  194.6/
  DATA subname( 83), dif0( 83), ar( 83), meso( 83), lebas( 83) / 'IO              ',0.1002,    8.0,      0.0,   44.4/
  DATA subname( 84), dif0( 84), ar( 84), meso( 84), lebas( 84) / 'OIO             ',0.0938,    8.0,      0.0,   51.8/
  DATA subname( 85), dif0( 85), ar( 85), meso( 85), lebas( 85) / 'I2O2            ',0.0732,    8.0,      0.0,   88.8/
  DATA subname( 86), dif0( 86), ar( 86), meso( 86), lebas( 86) / 'I2O3            ',0.0707,    8.0,      0.0,   96.2/
  DATA subname( 87), dif0( 87), ar( 87), meso( 87), lebas( 87) / 'I2O4            ',0.0684,    8.0,      0.0,  103.6/
  DATA subname( 88), dif0( 88), ar( 88), meso( 88), lebas( 88) / 'HI              ',0.1045,    8.0,      0.0,   40.7/
  DATA subname( 89), dif0( 89), ar( 89), meso( 89), lebas( 89) / 'HOI             ',0.0972,    8.0,      0.0,   48.1/
  DATA subname( 90), dif0( 90), ar( 90), meso( 90), lebas( 90) / 'INO             ',0.0882,    8.0,      0.0,   60.9/
  DATA subname( 91), dif0( 91), ar( 91), meso( 91), lebas( 91) / 'INO2            ',0.0883,   20.0,      0.0,   69.2/
  DATA subname( 92), dif0( 92), ar( 92), meso( 92), lebas( 92) / 'IONO2           ',0.0792,    8.0,      0.0,   77.5/
  DATA subname( 93), dif0( 93), ar( 93), meso( 93), lebas( 93) / 'BRO             ',0.1144,    1.0,      0.0,   34.4/
  DATA subname( 94), dif0( 94), ar( 94), meso( 94), lebas( 94) / 'HOBR            ',0.1101,    1.0,      0.0,   38.1/
  DATA subname( 95), dif0( 95), ar( 95), meso( 95), lebas( 95) / 'HBR             ',0.1216,    2.0,      0.0,   30.7/
  DATA subname( 96), dif0( 96), ar( 96), meso( 96), lebas( 96) / 'BRONO2          ',0.0855,    1.0,      0.0,   67.5/
  DATA subname( 97), dif0( 97), ar( 97), meso( 97), lebas( 97) / 'BRNO2           ',0.0909,    1.0,      0.0,   59.2/
  DATA subname( 98), dif0( 98), ar( 98), meso( 98), lebas( 98) / 'BRCL            ',0.0966,    1.0,      0.0,   51.6/
  DATA subname( 99), dif0( 99), ar( 99), meso( 99), lebas( 99) / 'DMS             ',0.0926,    2.0,      0.0,   77.4/
  DATA subname(100), dif0(100), ar(100), meso(100), lebas(100) / 'MSA             ',0.0896,    2.0,      0.0,   77.4/
  DATA subname(101), dif0(101), ar(101), meso(101), lebas(101) / 'METHANE         ',0.2107,    2.0,      0.0,   29.6/ ! dif0, equation 9-22. Scwarzenbach et. (1993) Env. Org. Chem.
  DATA subname(102), dif0(102), ar(102), meso(102), lebas(102) / 'ACRYACID        ',0.0908,    2.0,      0.0,   63.2/ 
  DATA subname(103), dif0(103), ar(103), meso(103), lebas(103) / 'CARBSULFIDE     ',0.1240,    5.0,      0.0,   51.5/ 
  DATA subname(104), dif0(104), ar(104), meso(104), lebas(104) / 'ACETONITRILE    ',0.1280,    5.0,      0.0,   52.3/ 
  DATA subname(105), dif0(105), ar(105), meso(105), lebas(105) / '6_NITRO_O_CRESOL',0.0664,   16.0,      0.0,  155.0/ ! dif0, equation 9-22. Scwarzenbach et. (1993) Env. Org. Chem.

  IF ( first_call ) THEN

     !$omp barrier
     !$omp single
     first_call = .FALSE.

     DO l = 1, n_spc_m3dry
        scc_pr_23( l ) = 0.2 * ( dif0( l ) * pr / kvis ) ** twothird
        ddif0    ( l ) = dif0( l ) / dwat
        csnow    ( l ) = ar( l ) / ( a0 * rsnow0 )
        cgnd     ( l ) = ar( l ) / ( a0 * rg0 ) 
        ccut     ( l ) = ar( l ) / ( a0 * rcut0 )
        dw25     ( l ) = 13.26e-5 / ( 0.8904**1.14 * lebas( l )**0.589 )
     END DO

     ! Set up Henry's law species indices

     hlspc( : ) = 0
     hleff( : ) = .false.

     do l = 1, n_spc_m3dry
        if ( use_depspc( l ) ) then
           call hlconst_spcs_init ( subname( l ), hlspc( l ) )
           hleff( l ) = is_effect_spc( hlspc( l ) )
        endif
     enddo
     !$omp end single

  END IF   ! first_call

!-------------------------------------------------------------------------------
! Loop over land cells and calculate dry deposition.
!-------------------------------------------------------------------------------

  effective = .TRUE.

  do jwl = 1, itab_w(iw)%nland
     iwl = itab_w(iw)%iland(jwl)

     kw  = itab_wl(iwl)%kw

     ustar = land%ustar(iwl)
     zf    = dzt(kw)
     rai   = land%ggaer(iwl)
     wtv0  = land%wthv(iwl)
     lai   = land%veg_lai(iwl)
     tai   = land%veg_tai(iwl)
     vegfr = land%veg_fracarea(iwl)
     wrcr  = land%veg_water(iwl)
     hveg  = land%veg_height(iwl)

     tempvg = land%veg_temp(iwl)
     tempcr = land%cantemp(iwl)
     rh_air = 100.0 * land%canshv(iwl) * real(rho(kw,iw)) / rhovsil( land%cantemp(iwl) - 273.15 )
     rh_air = min( 100.0, max( rh_air, 0.0 ) )

     ifurban = ( any(land%leaf_class(iwl) == (/ 19, 21 /)) )

     if (wtv0 > 0.0) then
        wstar  = (grav * pblh(iw) * wtv0 / theta(kw,iw)) ** onethird
     else
        wstar  = 0.0
     endif

     if ( land%flag_vg(iwl) ) then
        slmsts = slmsts_vg(land%ntext_soil(nzg,iwl))
     else
        slmsts = slmsts_ch(land%ntext_soil(nzg,iwl))
     endif

     slmst1 = min(land%soil_water(nzg,iwl), slmsts)

     if ( land%nlev_sfcwater(iwl) > 0 ) then

        ! Surface wetness factor
        deltag = min( 1.0, (sum(land%sfcwater_mass(1:land%nlev_sfcwater(iwl),iwl)) / 0.2)**2 ) ** onethird

        ! Determine sfcwater temperature and liquid water mass fraction
        call qtk( land%sfcwater_energy(land%nlev_sfcwater(iwl),iwl), tempgr, liqfrg )

        tai  = tai  * (1.0 - land%snowfac(iwl))
        lai  = lai  * (1.0 - land%snowfac(iwl))
        hveg = hveg * (1.0 - land%snowfac(iwl))

     else

        liqfrg  = 0.0
        deltag  = 0.0

        ! Determine the bare ground temperature
        call qwtk( land%soil_energy(nzg,iwl),land%soil_water(nzg,iwl)*1.e3, &
                   slcpd(land%ntext_soil(nzg,iwl)), tempgr, xm )
     endif

     ! Canopy Wetness

     wrmax  = 0.2 * max(tai,1.e-10)
     deltav = min( 1.0, (wrcr / wrmax)**2 ) ** onethird
     liqfrv = merge(0.0, 1.0, tempvg < 273.16)

     ! PH of canopy/ground water

     if (ccnparm < 1.e6) then

        fact = 0.05 * (cldnum(iw) * 1.e-7 - 10.0)
        fact = min( max(fact, 0.0), 1.0 )
        hplus = fact * hplus_dirty + (1.0 - fact) * hplus_clean

     else

        hplus = hplus_def

     endif

     ! Vegetation properties

     if (lai > 0.01 .and. vegfr > 0.01) then

        laiv = lai / vegfr
        taiv = tai / vegfr

        ! Stomatal conductance

        rsi = 1.0 / land%stom_resist(iwl)

        ! Canopy resisance

        rinc = 14.0 * taiv * hveg / ustar

     endif

     IF ( sfc_hono ) THEN

! Calculate A/V for leaves.
! LAI was multiplied by 2 to account for the fact that surface area
! is provided by both sides of the leaves.
! Matthews Jones, Ammonia deposition to semi-natural vegetation,
! PhD dissertation, University of Dundee, Scotland, 2006

        surf_leaf = 2.0 * tai / zf

! Calculate A/V for buildings and other structures.
! Buildings and other structures can provide additional surfaces in
! urban areas for the heterogeneous reaction to occur. However, such
! information is not readily available; in the absence of such information,
! it is scaled to purb(c,r). Svensson et al., (1987) suggests a typical value
! of 0.2 for A/V for buildings in urban environments. A maximum value of 0.2
! for A/V for buildings is assigned to the grid cell containing the highest
! purb(c,r) i.e., 100.0. A/V for buildings for other grid-cell is calculated
! as purb(c,r)*(0.2/100.0); Cai et al. (2006) used a value of 1.0 for their
! study at New York (total A/V)

        if ( ifurban ) then
           surf_bldg = 0.2
        else
           surf_bldg = 0.0
        endif

! Calculate rate constant for the reaction (psudeo-first order reaction,
! unit per second). Calculate pseudo-first order rate constant using Eq 1
! of Vogel et al. (2003).  Unit of KNO2 is in 1/min in the paper; divide it
! by 60 to convert it into 1/sec.

        kno2 = MAX( 0.0, 5.0E-5 * (surf_leaf + surf_bldg) )

     endif

     ! Loop over species to calculate dry deposition velocities.

     depvel_gas_land( iwl,: ) = 0.0  ! initialize for this time period

     n = 0
     dloopl: DO l = 1, n_spc_m3dry
         
        IF ( .NOT. use_depspc( l ) ) CYCLE dloopl

        n = n + 1

        IF ( l .EQ. l_sulf ) THEN  ! Sulfate (SULF)

           ! Sulfate calculation follows Wesely (1985), Eqn. 11.

           rbsulfi = 0.002 * (ustar**2 + 0.24 * wstar**2) / ustar
           depvel_gas_land( iwl,n ) =  rai * rbsulfi / ( rai + rbsulfi )
               
        else

           ! Use CMAQ function for calculating the effective Henry's Law
           ! constant.  Note that original M3DRY wants inverse,
           ! non-dimensional Henry's Law (caq/cg).

           heff = hlconst( tempcr, effective, hplus, hlspc( l ) )

           if ( hleff( l ) .and. vegfr > 0.01 .and. lai > 0.01) then
              heff_ap = hlconst( tempcr, effective, hplus_ap, hlspc( l ) )
           else
              heff_ap = heff
           endif

           ! Make Henry's Law constant non-dimensional.
               
           heff  = heff * 0.08205 * tempcr

           ! Wet cuticle resistance.

           IF ( l .NE. l_o3 ) THEN
              rweti = heff * rcw0i
           ELSE
              ! Canopy level wet resistence Rwet to ozone was found to be about
              ! 200 s/m on basis of Keysburg exp. Using LAI(1-sided) of about
              ! 6.25 measured at Keysburg gives leaf level rwet about 1250 s/m
              ! Leaf level rwet from Altimir et al 2006 gives about 1350 s/m
              rweti = cweto3
           END IF

           ! wet (liquid) ground conductance

           IF ( l .Ne. l_O3 ) THEN
              rgliqi = heff * rgwet0i
           else
              rgliqi = 0.002  ! resistance of 500 s/m
           endif

           ! snow / frozen ground conductance

           rsnowi = csnow( l )

           ! wet ground conductance

           rgwi = (1.0 - liqfrg) * rsnowi + liqfrg * rgliqi

           ! dry ground conductance

           IF ( l .Ne. l_O3 ) THEN
              rgdi = cgnd( l )
           else
              rgdi = 1.0 / (200.0 + 300.0 * slmst1 / slmsts)
           endif

           ! Bare ground conductance

           rgndi  = ( 1.0 - deltag ) * rgdi + deltag * rgwi

           ! Vegetation calculations
           if (lai > 0.01 .and. vegfr > 0.01) then

              ! Ground conductance under canopy vegetation

              rgndci = rgndi / (1.0 + rgndi * rinc)

              ! Wet cuticle conductance with snow/ice

              rwetsfci = ( 1.0 - liqfrv) * rsnowi + liqfrv * rweti

              ! Dry cuticle conductance

              if ( l == l_nh3 ) then
                 rdryi = 0.00025 * exp( 0.054 * rh_air )
              elseif ( l == l_o3 ) then
                 rh_func = max( 0.0, (rh_air - 70.0) / 30.0 )
                 rdryi = ( 1.0 - rh_func ) * ccut( l ) + rh_func * rweti
              else
                 rdryi = ccut( l )
              endif

              ! Total cuticle conductance

              rcuti = taiv * ( (1.0 - deltav) * rdryi + deltav * rwetsfci )

              ! Bulk stomatal and mesophyll conductance (added in series)

              rstomi = ddif0( l ) * rsi
              rmesoi = heff_ap / 3000. + meso( l )

              rstomi = laiv * rmesoi * rstomi / (rstomi + rmesoi)

              ! Bulk surface conductance with vegetation (added in parallel)

              rci =  vegfr * (rstomi + rcuti + rgndci) + (1.0 - vegfr) * rgndi

           else

              rci = rgndi

           endif

           ! Compute dry deposition velocity (aerodynamic, surface, and laminar
           ! conductacnes added in series)

           rbci = ustar * scc_pr_23( l )

           depvel_gas_land( iwl,n ) = rci * rai * rbci / (rci*rai + rci * rbci + rai * rbci)

! HONO production via heterogeneous reaction on ground surfaces,
! 2NO2 = HONO + HNO3
! Rate constant for the reaction = (3.0E-3/60)* (A/V),
! where A/V is surface area/volume ratio
! HONO is produced and released into the atmosphere
! NO2 is lost via chemical reaction
! HNO3 is sticky and stays on the surfaces

! Determine NO2 concentration needed for HONO production term.

           IF ( sfc_hono .and. l .EQ. l_no2 ) THEN

! Loss of NO2 via the heterogeneous reaction is accounted as additional
! depositional loss. Add the loss of NO2 via the heterogeneous reaction
! to the regular deposition velocity (increased dep. vel.).  This will
! reduce the NO2 conc. in the atmosphere. Dep vel is adjusted back to the
! original value in vdiffacm2 after NO2 conc is reduced but before calculating
! depositional loss.

              depvel_gas_land( iwl,n ) = depvel_gas_land( iwl,n ) + 2.0 * kno2 * zf

! Calculate production (pvd) for HONO; unit = ppm * m/s
                     
              pvd = kno2 * cgrid(kw,iw,ic_no2) * zf
                     
! Add this to the HONO emissions for now when we apply it in vertical diffusion

              vdemis_gc(kw,iw,ie_hono) = vdemis_gc(kw,iw,ie_hono) &
                                       + pvd * land%area(iwl) * volti(kw,iw)

           ENDIF
               
        END IF   ! special condition for sulfate (SULF)

     END DO dloopl   ! (l = 1, n_spc_m3dry)

  enddo   ! land

!-------------------------------------------------------------------------------
! Loop over sea cells and calculate dry deposition.
!-------------------------------------------------------------------------------

  do jws = 1, itab_w(iw)%nsea
     iws = itab_w(iw)%isea(jws)

     kw  = itab_ws(iws)%kw

     isocean = (sea%leaf_class(iws) == 0)

     ustar  = sea%ustar(iws)
     ustarw = sea%sea_ustar(iws)
     rawi   = sea%sea_ggaer(iws)

     sst   = sea%seatc(iws)
     tw    = sea%sea_cantemp(iws) ! water surface film temperature
     wtv0  = sea%wthv(iws)

     if (wtv0 > 0.0) then
        wstar = (grav * pblh(iw) * wtv0 / theta(kw,iw)) ** onethird
     else
        wstar = 0.0
     endif

     kviswi = EXP( 0.025 * ( tw - stdtemp ) ) / 0.017

     if  (sea%nlev_seaice(iws) > 0 ) then
        ustari = sea%ice_ustar(iws)
        raicei = sea%ice_ggaer(iws)
        tice   = sea%ice_cantemp(iws) ! ice surface film temperature

        ! Determine the seaice liquid water mass fraction
        call qtk_sea( sea%seaice_energy(sea%nlev_seaice(iws),iws), xt, xm)
     endif

     ! Loop over species to calculate dry deposition velocities.

     depvel_gas_sea( iws,: ) = 0.0  ! initialize for this time period

     n = 0

     dloops: DO l = 1, n_spc_m3dry
         
        IF ( .NOT. use_depspc( l ) ) CYCLE dloops

        n = n + 1

        IF ( l .EQ. l_sulf ) THEN  ! Sulfate (SULF)

           ! Sulfate calculation follows Wesely (1985), Eqn. 11.

           rbsulfi = 0.002 * (ustar**2 + 0.24 * wstar**2) / ustar
           depvel_gas_sea( iws,n ) = rawi * rbsulfi / ( rawi + rbsulfi )

        else

           ! Use CMAQ function for calculating the effective Henry's Law
           ! constant.  Note that original M3DRY wants inverse, non-
           ! dimensional Henry's Law (caq/cg).   Water pH is different
           ! than rain, and we need to use the water temperature.

           heff  = hlconst( tw, effective, hplus_h2o, hlspc( l ) )

           ! Make Henry's Law constant non-dimensional.

           heff  = heff * 0.08205 * tw

           ! from Hayduk and Laudie

           dw        = dw25( l ) * ( tw * rt25inK ) * ( 0.009025 * kviswi )
           scw_pr_23 = d3 * ( (dw * pr * kviswi)**2 ) ** onethird

           if (l == l_o3) then  ! implement Chang et al(2004)
                  
              ! pChang is a/H or alpha/H which would be 1/H in current model
              ! note that in Chang et al (2004) and Garland et al (1980), their 
              ! H is Cair/Cwater wich is the inverse of heff

              pChang  = 1.75
              kwChang = ustarw * scw_pr_23

              ! If a file of chlorophyll concentrations is provided, Iodide concentration 
              ! are estimated from a fit to the Rebello et al 1990 data. The slope and 
              ! correlation are given in the paper but not the intercept, so the data in 
              ! Tables 3 & 4 were fit to get the relationship below. The regression gives 
              ! the concentration in umol/L and is converted to mol/L for use in 
              ! Chang et al eq. The slope and correlation are a slightly different than
              ! in Table 5. If chlorophyll concs are not available, a constant value 
              ! for [I-] of 100e-9 mol/l is used Use ocean file variables to determine if 
              ! the water cell is ocean or lake; method is only for ocean cells

              if (isocean) then
                 ! Iodide in sea-water based on SST  (mol /dm-3)
                 ciodide = 1.46E6 * EXP( -9134.0 / sst )
                 qiodide = sqrt ( 2.0e9 * ciodide * dw * 1.e-4 ) * heff
              else
                 qiodide = 0.0
              endif

              rsurfi = pChang * kwchang + qiodide

           else  ! not ozone

              rsurfi = heff * ustarw * scw_pr_23

           endif

           ! Compute dry deposition velocity over sea
           
           rbci = ustarw * scc_pr_23( l )

           depvel_gas_sea( iws,n ) = rsurfi * rawi * rbci / ( rsurfi * rawi + rsurfi * rbci + rawi * rbci )

           ! Include fractional contribution of seaice if present

           if  (sea%nlev_seaice(iws) > 0 ) then

              ! Use CMAQ function for calculating the effective Henry's Law
              ! constant.  Note that original M3DRY wants inverse,
              ! non-dimensional Henry's Law (caq/cg).

              heff  = hlconst( tice, effective, hplus_h2o, hlspc( l ) )

              ! Make Henry's Law constant non-dimensional

              heff  = heff * 0.08205 * tw

              ! Wet surface resistance

              rgwi  = rgwet0i * heff

              ! Assume seaice behaves similarly to snowcover

              rsurfi = (1.0 - xm) * csnow(l) + xm * rgwi

              ! Compute dry deposition velocity contribution from seaice

              rbci  = ustari * scc_pr_23( l )

              d_ice = rsurfi * raicei * rbci / ( rsurfi * raicei + rsurfi * rbci + raicei * rbci )

              depvel_gas_sea( iws,n ) = (1.0 - sea%seaicec(iws)) * depvel_gas_sea( iws,n ) &
                                      +        sea%seaicec(iws)  * d_ice

           endif  ! seaice
               
        endif  ! special condition for sulfate (SULF)

     END DO dloops   ! (l = 1, n_spc_m3dry)

  enddo   ! sea

END SUBROUTINE m3dry
