!===============================================================================
! OLAM was originally developed at Duke University by Robert Walko, Martin Otte,
! and David Medvigy in the project group headed by Roni Avissar.  Development
! has continued by the same team working at other institutions (University of
! Miami (rwalko@rsmas.miami.edu), the Environmental Protection Agency, and
! Princeton University), with significant contributions from other people.

! Portions of this software are copied or derived from the RAMS software
! package.  The following copyright notice pertains to RAMS and its derivatives,
! including OLAM:  

   !----------------------------------------------------------------------------
   ! Copyright (C) 1991-2006  ; All Rights Reserved ; Colorado State University; 
   ! Colorado State University Research Foundation ; ATMET, LLC 

   ! This software is free software; you can redistribute it and/or modify it 
   ! under the terms of the GNU General Public License as published by the Free
   ! Software Foundation; either version 2 of the License, or (at your option)
   ! any later version. 

   ! This software is distributed in the hope that it will be useful, but
   ! WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   ! or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   ! for more details.
 
   ! You should have received a copy of the GNU General Public License along
   ! with this program; if not, write to the Free Software Foundation, Inc.,
   ! 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA 
   ! (http://www.gnu.org/licenses/gpl.html) 
   !----------------------------------------------------------------------------

!===============================================================================

subroutine leaf4()

  use leaf_coms,      only: nzs, iupdndvi, s1900_ndvi, dt_leaf, &
                            indvifile, nndvifiles, soil_rough, snow_rough, z_root
  use mem_ijtabs,     only: itabg_w, itab_w
  use mem_sfcg,       only: itab_wsfc, itab_vsfc, sfcg, mvsfc, mwsfc
  use mem_land,       only: land, mland, omland, nzg, slzt, dslz
  use mem_lake,       only: lake, mlake, omlake
  use mem_sea,        only: sea, omsea
  use misc_coms,      only: io6, s1900_sim, isubdomain, time8p, iparallel
  use mem_para,       only: myrank
  use mem_basic,      only: rho, press, theta, tair, vxe, vye, vze, sh_v
  use mem_micro,      only: sh_c
  use consts_coms,    only: grav, cliq1000, alli1000
  use leaf4_canopy,   only: canopy, vegndvi
  use leaf4_sfcwater, only: sfcwater, sfcwater_adjust, remove_runoff
  use leaf4_soil,     only: soil, soil_wat2khyd
  use leaf4_plot,     only: leaf_plot
  use therm_lib,      only: qwtk, qtk
  use olam_mpi_sfcg,  only: mpi_send_wsfc, mpi_recv_wsfc

  implicit none

  ! Local arrays

  real :: sfcwater_tempk   (nzs) ! surface water temperature [K]
  real :: sfcwater_fracliq (nzs) ! fraction of sfc water in liquid phase
  real :: soil_tempk       (nzg) ! soil temperature [K]
  real :: soil_fracliq     (nzg) ! fraction of soil moisture in liquid phase
  real :: energy_per_m2    (nzs) ! sfcwater energy [J/m^2]
  real :: hydresist1       (nzg)
  real :: hydresist2       (nzg)
  real :: dheight          (nzg) ! change in water height (of a lake surface or in a soil layer)
                                 ! from lateral fluxes [m]
  real :: energyin         (nzg) ! change in energy (of lake water or a soil layer) from lateral fluxes [J/m^2]

  real :: head        (nzg,mland)
  real :: head_slope  (nzg,mland)
  real :: soil_watfrac(nzg,mland)
  real :: watflux     (nzg,mvsfc) ! [m/s]
  real :: energyflux  (nzg,mvsfc)

  real, parameter :: timescale_lake = 7200. ! Lake cell-to-cell equilibration time scale [s]

  ! Local variables

  integer :: iland, ilandn, j, iwsfc, ilake
  integer :: ivn, iwn, ivsfc
  integer :: iw1, iw2, iland1, iland2, ilake1, ilake2, isea1, isea2

  integer :: k     ! vertical index over soil layers
  integer :: nlsw1 ! maximum of (1,land%nlev_sfcwater(iland))
  integer :: icomb ! implicit heat balance flag [0=no, 1=yes]

  integer :: ktrans ! vertical index of soil layer supplying transpiration

  real :: transp  ! transpiration xfer this LEAF timestep [kg/m^2]
  real :: wfree1  ! sfcwater bottom layer free water mass [kg/m^2]
  real :: qwfree1 ! sfcwater bottom layer free water energy [J/m^2]
  real :: dwfree1 ! sfcwater bottom layer free water depth [m]

  real :: timefac_ndvi  ! fraction of elapsed time from past to future NDVI obs
  real :: dnvo2, dirv, soil_watfracv

  real :: khyd1, khyd2
  real :: factor

  real, parameter :: thermcond_dry_organic = 0.05  ! [W/(m K)] 
  real, parameter :: thermcond_sat_organic = 0.25  ! [W/(m K)] 
  real, parameter :: thermcond_bedrock     = 3.0   ! [W/(m K)] 
  real, parameter :: thermcond_liq         = 0.6   ! [W/(m K)] 
  real, parameter :: thermcond_ice         = 2.29  ! [W/(m K)] 

  real :: thermcond_dry_mineral
  real :: thermcond_sat_mineral

  real :: mineral
  real :: bulkdens_mineral
  real :: thermcond_drysoil
  real :: thermcond_sat_solids
  real :: thermcond_sat_soil
  real :: kersten_liq, kersten_ice
  real :: kersten
  real :: thermcond_soil(nzg)

  integer, parameter :: iland_print = 0
  real, parameter :: snowmin_expl = 10. ! min sfcwater mass for explicit heat
                                        ! xfer [kg/m^2]

  real, external :: rhovsil ! function to compute sat spec hum over liquid or ice

  !=================================================================
  ! TEMPORARY RUNOFF VARIABLES
  real :: runoff
  real :: qrunoff
  !=================================================================

  !---------------------------------------------------------------------------
  ! Compute hydraulic head in primary land cells and those adjacent to primary
  ! cells, and compute horizontal groundwater fluxes of mass and energy across
  ! faces of all primary land and lake cells with IVORONOI = 3 in this subdomain
  !---------------------------------------------------------------------------

  watflux   (:,:) = 0.
  energyflux(:,:) = 0.

  ! Loop over all surface cells (in this subdomain)

  !$omp do private(iland)
  do iwsfc = 2,mwsfc

     ! Skip this cell if running in parallel and cell rank is not MYRANK

     if (isubdomain == 1 .and. itab_wsfc(iwsfc)%irank /= myrank) cycle

     ! Skip this cell if it is not land

     if (sfcg%leaf_class(iwsfc) < 2) cycle

     ! Compute total head (relative to local topographic datum) based on soil
     ! moisture, pressure head, and depth in soil

     iland = iwsfc - omland

     call head_column(nzg, iland, slzt(:),       &
                      land%soil_water (:,iland), &
                      land%wresid_vg  (:,iland), &
                      land%wsat_vg    (:,iland), &
                      land%alpha_vg   (:,iland), &
                      land%en_vg      (:,iland), &
                      land%head_press (:,iland), &
                      head            (:,iland), &
                      head_slope      (:,iland), &
                      soil_watfrac    (:,iland)  )

  enddo ! iwsfc
  !$omp end do

  ! MPI send/recv of time-dependent SFCG, LAND, LAKE, SEA quantities

  if (iparallel == 1) then
     call mpi_send_wsfc(head=head,soil_watfrac=soil_watfrac)
     call mpi_recv_wsfc(head=head,soil_watfrac=soil_watfrac)
  endif

  ! Loop over all SFC grid V faces in this subdomain (NOTE: they only exist
  ! around Voronoi cells).  Evaluate water and energy fluxes across V face,
  ! with flux direction defined by flux = (head(v%iw1) - head(v%iw2)) / resist

  !$omp parallel private(soil_tempk, soil_fracliq, hydresist1, hydresist2)
  !$omp do private(iw1, iw2, iland1, iland2, k, ilake1, ilake2, isea1, isea2, &
  !$omp            dnvo2, soil_watfracv, khyd1, khyd2)
  do ivsfc = 2,mvsfc
     iw1 = itab_vsfc(ivsfc)%iwn(1)
     iw2 = itab_vsfc(ivsfc)%iwn(2)

     ! Skip this SFC grid V face if either adjacent W cell is not a Voronoi cell

     if (iw1 < 2 .or. iw2 < 2) cycle
     if (itab_wsfc(iw1)%ivoronoi < 2 .or. itab_wsfc(iw2)%ivoronoi < 2) cycle

        ! Compute half distance along topographic slope between adjacent sfcg w points

        dnvo2 = 0.5 * sqrt(sfcg%dnv(ivsfc)**2 + (sfcg%topw(iw1) - sfcg%topw(iw2))**2)

     if (sfcg%leaf_class(iw1) >= 2 .and. sfcg%leaf_class(iw2) >= 2) then

        ! Fluxes at current IVSFC location are between two land cells

        iland1 = iw1 - omland
        iland2 = iw2 - omland

        ! Loop over vertical levels

        do k = 1,nzg

           ! Compute soil water fraction at ivsfc and hydraulic resistance on
           ! iw1 and iw2 sides of ivsfc face

           soil_watfracv = .5 * (soil_watfrac(k,iland1) + soil_watfrac(k,iland2))

           call soil_wat2khyd(soil_watfracv, land%ksat_vg(k,iland1), land%lambda_vg(k,iland1), &
                              land%en_vg(k,iland1), khyd1, land%gpp(iland1), slzt(k),          &
                              z_root(sfcg%leaf_class(iw1)), head(k,iland1))

           call soil_wat2khyd(soil_watfracv, land%ksat_vg(k,iland2), land%lambda_vg(k,iland2), &
                              land%en_vg(k,iland2), khyd2, land%gpp(iland2), slzt(k),          &
                              z_root(sfcg%leaf_class(iw2)), head(k,iland2))

           ! Compute hydraulic resistances between iw1 and iw2 cells

           hydresist1(k) = dnvo2 / khyd1
           hydresist2(k) = dnvo2 / khyd2

           ! Compute horizontal mass and energy fluxes across this cell face,
           ! taking into account the topographic gradient since it is not
           ! included in head(:,:).  Modulate fluxes if water is partially or
           ! completely frozen.

           watflux(k,ivsfc) = (head(k,iland1) + sfcg%topw(iw1) - head(k,iland2) - sfcg%topw(iw2)) &
                            / (hydresist1(k) + hydresist2(k)) ! [m/s]

           if (watflux(k,ivsfc) > 0.) then

              call qwtk(land%soil_energy(k,iland1),land%soil_water(k,iland1)*1.e3, &
                        land%specifheat_drysoil(k,iland1),soil_tempk(k),soil_fracliq(k))

              watflux(k,ivsfc) = watflux(k,ivsfc) * soil_fracliq(k)

              energyflux(k,ivsfc) = watflux(k,ivsfc) &
                                  * (cliq1000 * (soil_tempk(k) - 273.15) + alli1000) ! [J/(m^2 s)]

           else

              call qwtk(land%soil_energy(k,iland2),land%soil_water(k,iland2)*1.e3, &
                        land%specifheat_drysoil(k,iland2),soil_tempk(k),soil_fracliq(k))

              watflux(k,ivsfc) = watflux(k,ivsfc) * soil_fracliq(k)

              energyflux(k,ivsfc) = watflux(k,ivsfc) &
                                  * (cliq1000 * (soil_tempk(k) - 273.15) + alli1000) ! [J/(m^2 s)]
           endif

        enddo

     elseif (sfcg%leaf_class(iw1) >= 2) then

        ! Fluxes at current IVSFC location are between a land cell (iw1) and
        ! a water cell (iw2) - either sea or lake.  Hydraulic resistance is
        ! from land side only with assumed saturation at interface.  Head is
        ! the same at all levels in water cells (relative to surface elevation).

        iland1 = iw1 - omland

        ! Loop over vertical levels

        do k = 1,nzg

           ! Compute soil water fraction at ivsfc and hydraulic resistance on
           ! iw1 side of ivsfc face

           soil_watfracv = .5 * (soil_watfrac(k,iland1) + 1.0)

           ! Compute hydraulic resistances between iw1 and iw2 cells

           call soil_wat2khyd(soil_watfracv, land%ksat_vg(k,iland1), land%lambda_vg(k,iland1), &
                              land%en_vg(k,iland1), khyd1, land%gpp(iland1), slzt(k),          &
                              z_root(sfcg%leaf_class(iw1)), head(k,iland1))

           ! Compute hydraulic resistances between iw1 and iw2 cells

           hydresist1(k) = dnvo2 / khyd1

           ! Compute horizontal mass and energy fluxes across this cell face,
           ! taking into account the topographic gradient since it is not
           ! included in head(:,:).  Modulate fluxes if water is partially
           ! or completely frozen.

           watflux(k,ivsfc) = (head(k,iland1) + sfcg%topw(iw1) - sfcg%head1(iw2) - sfcg%topw(iw2)) &
                            / hydresist1(k) ! [m/s]

           if (watflux(k,ivsfc) > 0.) then

              call qwtk(land%soil_energy(k,iland1),land%soil_water(k,iland1)*1.e3, &
                        land%specifheat_drysoil(k,iland1),soil_tempk(k),soil_fracliq(k))

              watflux(k,ivsfc) = watflux(k,ivsfc) * soil_fracliq(k)

              energyflux(k,ivsfc) = watflux(k,ivsfc) &
                                  * (cliq1000 * (soil_tempk(k) - 273.15) + alli1000) ! [J/(m^2 s)]

           else

              if (sfcg%leaf_class(iw2) == 0) then
                 isea2 = iw2 - omsea
                 energyflux(k,ivsfc) = watflux(k,ivsfc) &
                                     * (cliq1000 * (sea%seatc(isea2) - 273.15) + alli1000) ! [J/(m^2 s)]
              else
                 ilake2 = iw2 - omlake
                 energyflux(k,ivsfc) = watflux(k,ivsfc) &
                                     * (cliq1000 * (lake%sfctemp(ilake2) - 273.15) + alli1000) ! [J/(m^2 s)]
              endif

           endif

        enddo ! k

     elseif (sfcg%leaf_class(iw2) >= 2) then

        ! Fluxes at current IVSFC location are between a land cell (iw2) and
        ! a water cell (iw1) - either sea or lake.  Hydraulic resistance is
        ! from land side only with assumed saturation at interface.  Head is
        ! the same at all levels in water cells (relative to surface elevation).

        iland2 = iw2 - omland

        ! Loop over vertical levels

        do k = 1,nzg

           ! Compute soil water fraction at ivsfc and hydraulic resistance on
           ! iw2 side of ivsfc face

           soil_watfracv = .5 * (soil_watfrac(k,iland2) + 1.0)

           call soil_wat2khyd(soil_watfracv, land%ksat_vg(k,iland2), land%lambda_vg(k,iland2), &
                              land%en_vg(k,iland2), khyd2, land%gpp(iland2), slzt(k),          &
                              z_root(sfcg%leaf_class(iw2)), head(k,iland2))

           ! Compute hydraulic resistances between iw1 and iw2 cells

           hydresist2(k) = dnvo2 / khyd2

           ! Compute horizontal mass and energy fluxes across this cell face,
           ! taking into account the topographic gradient since it is not
           ! included in head(:,:).  Modulate fluxes if water is partially 
           ! or completely frozen.

           watflux(k,ivsfc) = (sfcg%head1(iw1) + sfcg%topw(iw1) - head(k,iland2) - sfcg%topw(iw2)) &
                            / hydresist2(k) ! [m/s]

           if (watflux(k,ivsfc) > 0.) then

              if (sfcg%leaf_class(iw1) == 0) then
                 isea1 = iw1 - omsea
                 energyflux(k,ivsfc) = watflux(k,ivsfc) &
                                     * (cliq1000 * (sea%seatc(isea1) - 273.15) + alli1000) ! [J/(m^2 s)]
              else
                 ilake1 = iw1 - omlake
                 energyflux(k,ivsfc) = watflux(k,ivsfc) &
                                     * (cliq1000 * (lake%sfctemp(ilake1) - 273.15) + alli1000) ! [J/(m^2 s)]
              endif

           else

              call qwtk(land%soil_energy(k,iland2),land%soil_water(k,iland2)*1.e3, &
                        land%specifheat_drysoil(k,iland2),soil_tempk(k),soil_fracliq(k))

              watflux(k,ivsfc) = watflux(k,ivsfc) * soil_fracliq(k)

              energyflux(k,ivsfc) = watflux(k,ivsfc) &
                                  * (cliq1000 * (soil_tempk(k) - 273.15) + alli1000) ! [J/(m^2 s)]

           endif

        enddo ! k

     elseif (sfcg%leaf_class(iw1) == 1 .and. sfcg%leaf_class(iw2) == 1) then

        ! Fluxes at current IVSFC location are between two lake cells.  If they
        ! have different topographic heights, assume that they are unconnected
        ! and skip this IVN point.

        if (abs(sfcg%topw(iw1) - sfcg%topw(iw2)) > 0.1) cycle

        ilake1 = iw1 - omlake
        ilake2 = iw2 - omlake

        ! Given their (nearly) identical topographic heights, the two lake
        ! cells are assumed to be interconnected.  Compute a water flux at
        ! ivsfc that is proportional to the difference in head1 values between
        ! the lake cells so that water levels tend toward equilibration.

        watflux(nzg,ivsfc) = (sfcg%head1(iw1) + sfcg%topw(iw1) - sfcg%head1(iw2) - sfcg%topw(iw2)) &
                           * min(sfcg%area(iw1),sfcg%area(iw2)) &
                           / (sfcg%dnu(ivsfc) * dslz(nzg) * timescale_lake)  ! [m/s]

        if (watflux(nzg,ivsfc) > 0.) then

           energyflux(nzg,ivsfc) = watflux(nzg,ivsfc) &
                                 * (cliq1000 * (lake%sfctemp(ilake1) - 273.15) + alli1000) ! [J/(m^2 s)]
        else

           energyflux(nzg,ivsfc) = watflux(nzg,ivsfc) &
                                 * (cliq1000 * (lake%sfctemp(ilake2) - 273.15) + alli1000) ! [J/(m^2 s)]

        endif

     endif ! sfcg%leaf_class(iw1 & iw2)

  enddo ! ivsfc
  !$omp end do
  !$omp end parallel

  ! Time interpolation factor for updating NDVI

  timefac_ndvi = 0.

  if (iupdndvi == 1 .and. nndvifiles > 1) then
     timefac_ndvi = (s1900_sim             - s1900_ndvi(indvifile-1))  &
                  / (s1900_ndvi(indvifile) - s1900_ndvi(indvifile-1))
  endif

  !-----------------------------------------------------------------------
  ! Loop over ALL LAKE CELLS
  !-----------------------------------------------------------------------

  !$omp parallel private(dheight, energyin)
  !$omp do private(iwsfc, j, ivn, iwn, dirv, k, factor)
  do ilake = 2,mlake
     iwsfc = ilake + omlake

     ! Skip this cell if running in parallel and cell rank is not MYRANK

     if (isubdomain == 1 .and. itab_wsfc(iwsfc)%irank /= myrank) cycle

     ! Skip this cell if it is not interior voronoi cell, which indicates that
     ! horizontal water fluxes do not participate in its water budget

     if (itab_wsfc(iwsfc)%ivoronoi /= 3) cycle

     dheight (:) = 0.
     energyin(:) = 0.

     ! Loop over all lateral faces of this ilake cell

     do j = 1,itab_wsfc(iwsfc)%npoly
        ivn = itab_wsfc(iwsfc)%ivn(j)
        iwn = itab_wsfc(iwsfc)%iwn(j)

        if (iwsfc == itab_vsfc(ivn)%iwn(2)) then
           dirv = 1.0
        else
           dirv = -1.0
        endif

        if (sfcg%leaf_class(iwn) >= 2) then

           ! Neighbor cell is land; loop over its vertical levels and sum
           ! mass and energy fluxes.  Accumulate sum only in nzg level.

           do k = 1,nzg
              factor = dirv * dt_leaf * sfcg%dnu(ivn) * dslz(k) / sfcg%area(iwsfc) ! [s/m]

              dheight (nzg) = dheight (nzg) + factor * watflux   (k,ivn) ! [m]
              energyin(nzg) = energyin(nzg) + factor * energyflux(k,ivn) ! [J/m^2]
           enddo

        elseif (sfcg%leaf_class(iwn) == 1) then

           ! Neighbor cell is lake; only top-level flux is nonzero.

           factor = dirv * dt_leaf * sfcg%dnu(ivn) * dslz(nzg) / sfcg%area(iwsfc) ! [s/m]

           dheight (nzg) = dheight (nzg) + factor * watflux   (nzg,ivn) ! [m]
           energyin(nzg) = energyin(nzg) + factor * energyflux(nzg,ivn) ! [J/m^2]

        endif

     enddo ! j

     ! Add height and energy changes to cell

     sfcg%head1(iwsfc) = sfcg%head1(iwsfc) + dheight(nzg)

     lake%sfctemp(ilake) = lake%sfctemp(ilake) + energyin(nzg) / (lake%depth(ilake) * cliq1000)

  enddo
  !$omp end do
  !$omp end parallel

  !-----------------------------------------------------------------------
  ! Loop over ALL LAND CELLS
  !-----------------------------------------------------------------------

  !$omp parallel private(dheight, energyin, soil_tempk, soil_fracliq, &
  !$omp                  thermcond_soil, sfcwater_tempk, sfcwater_fracliq, &
  !$omp                  energy_per_m2)

  !$omp do private(iwsfc, j, ivn, dirv, k, factor, icomb, nlsw1, &
  !$omp            bulkdens_mineral, mineral, thermcond_dry_mineral, &
  !$omp            thermcond_drysoil, thermcond_sat_mineral, &
  !$omp            thermcond_sat_solids, thermcond_sat_soil, &
  !$omp            kersten_liq, kersten_ice, kersten, &
  !$omp            ktrans, transp, wfree1, qwfree1, dwfree1)
  do iland = 2,mland
     iwsfc = iland + omland

     ! Skip this cell if running in parallel and cell rank is not MYRANK
     if (isubdomain == 1 .and. itab_wsfc(iwsfc)%irank /= myrank) cycle

     ! If this is interior Voronoi cell, apply horizontal groundwater mass
     ! and energy fluxes
 
     if (itab_wsfc(iwsfc)%ivoronoi == 3) then

        dheight (:) = 0.
        energyin(:) = 0.

        ! Loop over all lateral faces of this iland cell

        do j = 1,itab_wsfc(iwsfc)%npoly
           ivn = itab_wsfc(iwsfc)%ivn(j)

           if (iwsfc == itab_vsfc(ivn)%iwn(2)) then
              dirv = 1.0
           else
              dirv = -1.0
           endif

           ! Loop over vertical levels; sum mass and energy fluxes over j faces

           do k = 1,nzg
              factor = dirv * dt_leaf * sfcg%dnu(ivn) * dslz(k) / sfcg%area(iwsfc) ! [s]

              dheight (k) = dheight (k) + factor * watflux   (k,ivn) ! [m]
              energyin(k) = energyin(k) + factor * energyflux(k,ivn) ! [J/m^2]
           enddo ! k

        enddo ! j

        ! Add mass and energy changes to cell, and adjust hydraulic head due to
        ! water mass change according to constant slope fit for current timestep

        do k = 1,nzg
           land%soil_water (k,iland) = land%soil_water (k,iland) + dheight(k)
           land%soil_energy(k,iland) = land%soil_energy(k,iland) + energyin(k)

           head(k,iland) = head(k,iland) + head_slope(k,iland) * dheight(k)
        enddo

     endif ! ivoronoi == 3

     if (land%ed_flag(iland) == 0) then

!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
! START OF OLD LEAF4_LANDCELL SECTION
!~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

     ! Initialize variables for this land cell

     icomb = 1  ! Default initialization
     nlsw1 = max(land%nlev_sfcwater(iland),1)

     ! Diagnose soil temperature, liquid fraction, and thermal conductivity

     do k = 1,nzg
        call qwtk(land%soil_energy(k,iland),land%soil_water(k,iland)*1.e3, &
           land%specifheat_drysoil(k,iland),soil_tempk(k),soil_fracliq(k))

        ! Check whether bedrock or soil (negative ph_soil has been set as a flag for bedrock)

        if (land%ph_soil(k,iland) < 0.) then

           thermcond_soil(k) = thermcond_bedrock  ! Assuming that porosity is accounted for

        else

           ! Bulk density of mineral part of soil

           ! (Note: bulkdens_mineral uses wsat_vg which in turn uses bulkdens_drysoil, which
           ! includes both mineral and organic components.  Is there a more direct way
           ! to get bulkdens_mineral from bulkdens_drysoil?  One idea is:

           ! bulkdens_mineral = land%bulkdens_drysoil(k,iland) * (1. - land%organ(k,iland))

           ! This formula would presumably be more accurate for peat soil, for example,
           ! but would it ultimately give better values for thermcond_dry_mineral and
           ! thermcond_drysoil?

           bulkdens_mineral = 2700. * (1. - land%wsat_vg(k,iland))

           ! Thermal conductivity for dry soil material

           mineral = 1. - land%organ(k,iland)

           thermcond_dry_mineral = (0.135 * bulkdens_mineral + 64.7) &
                                 / (2700. - 0.947 * bulkdens_mineral)

           thermcond_drysoil = mineral             * thermcond_dry_mineral &
                             + land%organ(k,iland) * thermcond_dry_organic

           ! Thermal conductivity for saturated soil material

           thermcond_sat_mineral = (8.80 * land%sand(k,iland) + 2.92 * land%clay(k,iland)) &
                                 / (land%sand(k,iland) + land%clay(k,iland))

           thermcond_sat_solids = mineral             * thermcond_sat_mineral &
                                + land%organ(k,iland) * thermcond_sat_organic 
           !--------------------------------------------------------------------------------

           ! For soil case, use porosity and fractions of sand, clay, and organic matter

           ! Compute thermcond_sat_soil as geometric mean of solids, liquid, and ice contributions

           thermcond_sat_soil = thermcond_sat_solids**(1.-land%wsat_vg(k,iland)) &
                              * thermcond_liq**(land%wsat_vg(k,iland)*soil_fracliq(k)) &
                              * thermcond_ice**(land%wsat_vg(k,iland)*(1.-soil_fracliq(k)))

           kersten_liq = max(0.,log(soil_watfrac(k,iland)) + 1.)
           kersten_ice = soil_watfrac(k,iland)
           kersten = kersten_liq * soil_fracliq(k) + kersten_ice * (1. - soil_fracliq(k))

           thermcond_soil(k) = kersten * thermcond_sat_soil &
                       + (1. - kersten) * thermcond_drysoil

        endif

     enddo


!  if (iland == 129192) print*, 'swat12 ',land%soil_water(nzg,iland),land%sfcwater_mass(1,iland)

     ! Loop over all possible sfcwater (a.k.a. snowcover) layers

     do k = 1,nzs

        ! Check (temporary) for consistency between nlev_sfcwater and 
        ! sfcwater_mass.  If inconsistency found, print message and stop.

        if (land%sfcwater_mass(k,iland) > 1.e-12 .and. k > land%nlev_sfcwater(iland)) then
           write(io6,*) 'sfcwater mass is present in layer above nlev_sfcwater'
           write(io6,*) 'iland,k,glatw,glonw = ',iland,k, sfcg%glatw(iwsfc), sfcg%glonw(iwsfc)
           write(io6,*) 'nlev_sfcwater,sfcwater_mass(k) = ', &
                        land%nlev_sfcwater(iland),land%sfcwater_mass(k,iland)
           stop 'stop1 leaf4'
        endif

        if (land%sfcwater_mass(k,iland) < 1.e-6 .and. k <= land%nlev_sfcwater(iland)) then
           write(io6,*) 'sfcwater mass is absent in layer at or below nlev_sfcwater'
           write(io6,*) 'iland,k,glatw,glonw =',iland,k, sfcg%glatw(iwsfc), sfcg%glonw(iwsfc)
           write(io6,*) 'nlev_sfcwater,sfcwater_mass(k) = ', &
                        land%nlev_sfcwater(iland),land%sfcwater_mass(k,iland)
           stop 'stop2 leaf4'
        endif

        if (k <= land%nlev_sfcwater(iland)) then

           ! Diagnose sfcwater temperature, liquid fraction, and energy per m^2

           call qtk(land%sfcwater_energy(k,iland),sfcwater_tempk(k),sfcwater_fracliq(k))

           energy_per_m2(k) = land%sfcwater_energy(k,iland) * land%sfcwater_mass(k,iland)

        else

           ! Assign default values above active sfcwater layers

           land%sfcwater_mass  (k,iland) = 0.
           land%sfcwater_energy(k,iland) = 0.
           land%sfcwater_depth (k,iland) = 0.

           sfcwater_tempk  (k) = 273.15
           sfcwater_fracliq(k) = 0.
           energy_per_m2   (k) = 0.

        endif
     enddo


!  if (iland == 129192) print*, 'swat13 ',land%soil_water(nzg,iland),land%sfcwater_mass(1,iland)

     ! Update vegetation TAI, LAI, fractional area, albedo, and roughness

     call vegndvi(iland, iwsfc, timefac_ndvi, &
                  sfcg%leaf_class  (iwsfc), &
                  land%veg_height  (iland), &
                  land%veg_rough   (iland), &
                  land%veg_ndvip   (iland), &
                  land%veg_ndvif   (iland), &
                  land%veg_ndvic   (iland), &
                  land%veg_tai     (iland), &
                  land%veg_lai     (iland), &
                  land%veg_fracarea(iland), &
                  land%veg_albedo  (iland)  )

     ! Compute roughness length based on vegetation and snow.

     sfcg%rough(iwsfc) = max(soil_rough,land%veg_rough(iland)) * (1. - land%snowfac(iland)) + snow_rough

     ! Evaluate turbulent exchanges of heat and moisture between vegetation and canopy air
     ! and also between soil or snow surface and canopy air.  Evaluate transfer of
     ! precipitation moisture and heat to vegetation and shed from vegetation to surface.
     ! Update vegetation and canopy air temperatures resulting from these 
     ! plus radiative fluxes.


!if (iland == 29298) print*, 'le3 ',land%soil_water(nzg,iland),dheight(nzg)


!  if (iland == 129192) print*, 'swat14 ',land%soil_water(nzg,iland),land%sfcwater_mass(1,iland)

     call canopy(iland, iwsfc, nlsw1, icomb, ktrans, transp, snowmin_expl, &
                 sfcg%leaf_class           (iwsfc), &
                 sfcg%can_depth            (iwsfc), &
                 sfcg%rhos                 (iwsfc), &
                 sfcg%vels                 (iwsfc), &
                 sfcg%ustar                (iwsfc), &
                 sfcg%sxfer_t              (iwsfc), &
                 sfcg%sxfer_r              (iwsfc), &
                 sfcg%pcpg                 (iwsfc), &
                 sfcg%qpcpg                (iwsfc), &
                 sfcg%dpcpg                (iwsfc), &
                 sfcg%rshort               (iwsfc), &
                 sfcg%cantemp              (iwsfc), &
                 sfcg%canshv               (iwsfc), &
                 sfcg%glatw                (iwsfc), &
                 sfcg%glonw                (iwsfc), & 
                 land%snowfac              (iland), &
                 land%vf                   (iland), &
                 land%stom_resist          (iland), &
                 land%veg_height           (iland), &
                 land%veg_rough            (iland), &
                 land%veg_tai              (iland), &
                 land%veg_lai              (iland), &
                 land%hcapveg              (iland), &
                 land%rshort_v             (iland), &
                 land%rshort_g             (iland), &
                 land%rlong_v              (iland), &
                 land%rlong_s              (iland), &
                 land%rlong_g              (iland), &
                 land%veg_water            (iland), &
                 land%veg_temp             (iland), &   
                 land%sfcwater_mass  (nlsw1,iland), &
                 land%sfcwater_energy(nlsw1,iland), &
                 land%sfcwater_depth (nlsw1,iland), &
                 energy_per_m2       (nlsw1),       &
                 sfcwater_tempk      (nlsw1),       &
                 sfcwater_fracliq    (nlsw1),       &
                 land%soil_water        (1:nzg,iland), &
                 land%soil_energy       (1:nzg,iland), &
                 land%wsat_vg           (1:nzg,iland), &
                 land%ksat_vg           (1:nzg,iland), &
                 land%specifheat_drysoil(1:nzg,iland), &
                 head                   (1:nzg,iland), &
                 soil_tempk             (1:nzg),       &
                 soil_fracliq           (1:nzg)        )

     ! CALL SFCWATER:
     !  1. Compute soil and sfcwater heat conductivities
     !  2. Compute surface heat flux for top layer of soil or sfcwater
     !  3. Evaluate internal and bottom fluxes for sfcwater
     !  4. Update sfcwater layer energies due to heat flux and solar radiation
     !  5. Evaluate melting and percolation of liquid through sfcwater layers


! if (iland == 29298) print*, 'le4 ',land%soil_water(nzg,iland),dheight(nzg)


!  if (iland == 129192) print*, 'swat15 ',land%soil_water(nzg,iland),land%sfcwater_mass(1,iland)

     call sfcwater(iland, iwsfc, icomb, wfree1, qwfree1, dwfree1, snowmin_expl, &
                   sfcg%head1                   (iwsfc), &
                   land%nlev_sfcwater           (iland), &
                   land%sfcwater_mass     (1:nzs,iland), &
                   land%sfcwater_energy   (1:nzs,iland), &
                   land%sfcwater_depth    (1:nzs,iland), &
                   sfcwater_tempk         (1:nzs),       &
                   sfcwater_fracliq       (1:nzs),       &
                   energy_per_m2          (1:nzs),       &
                   land%soil_water        (1:nzg,iland), &
                   land%soil_energy       (1:nzg,iland), &
                   land%specifheat_drysoil(1:nzg,iland), &
                   soil_watfrac           (1:nzg,iland), &
                   soil_tempk             (1:nzg),       &
                   soil_fracliq           (1:nzg),       &
                   thermcond_soil         (1:nzg)        )


! if (iland == 29298) print*, 'le5 ',land%soil_water(nzg,iland),dheight(nzg)


!  if (iland == 129192) print*, 'swat16 ',land%soil_water(nzg,iland),land%sfcwater_mass(1,iland)

     call soil(iland, iwsfc, ktrans, transp, wfree1, qwfree1, &
               sfcg%leaf_class          (iwsfc), &
               sfcg%glatw               (iwsfc), &
               sfcg%glonw               (iwsfc), &
               sfcg%head1               (iwsfc), &
               land%head0               (iland), &
               land%nlev_sfcwater       (iland), &
               land%gpp                 (iland), &
               land%sfcwater_mass     (1,iland), &
               land%sfcwater_energy   (1,iland), &
               land%sfcwater_depth    (1,iland), &
               energy_per_m2          (1),       &
               land%soil_water    (1:nzg,iland), &
               land%soil_energy   (1:nzg,iland), &
               land%wresid_vg     (1:nzg,iland), &
               land%wsat_vg       (1:nzg,iland), &
               land%ksat_vg       (1:nzg,iland), &
               land%lambda_vg     (1:nzg,iland), &
               land%en_vg         (1:nzg,iland), &
               land%alpha_vg      (1:nzg,iland), &
               land%head_press    (1:nzg,iland), &
               soil_watfrac       (1:nzg,iland), &
               head_slope         (1:nzg,iland), &
               head               (1:nzg,iland), &
               soil_tempk         (1:nzg),       &
               soil_fracliq       (1:nzg),       &
               thermcond_soil     (1:nzg)        )


! if (iland == 29298) print*, 'le6 ',land%soil_water(nzg,iland),dheight(nzg)

     ! Inventory sfcwater layer(s) and adjust thicknesses if more than one


!  if (iland == 129192) print*, 'swat17 ',land%soil_water(nzg,iland),land%sfcwater_mass(1,iland)

     if (nzs > 1) then
        call sfcwater_adjust(iland, iwsfc,                      &
                             sfcg%glatw                (iwsfc), &
                             sfcg%glonw                (iwsfc), &
                             land%nlev_sfcwater        (iland), &
                             land%sfcwater_mass  (1:nzs,iland), &
                             land%sfcwater_energy(1:nzs,iland), &
                             land%sfcwater_depth (1:nzs,iland), &
                             energy_per_m2       (1:nzs)        )
     endif

     ! Compute surface and ground vap mxrat for next timestep; put into surface_ssh 
     ! and ground_ssv.

     nlsw1 = max(land%nlev_sfcwater(iland),1)


!  if (iland == 129192) print*, 'swat18 ',land%soil_water(nzg,iland),land%sfcwater_mass(1,iland)

     call grndvap(iland,                              &
                  sfcg%rhos                  (iwsfc), &
                  sfcg%canshv                (iwsfc), &
                  land%nlev_sfcwater         (iland), &
                  land%surface_ssh           (iland), &
                  land%ground_shv            (iland), &
                  land%sfcwater_energy (nlsw1,iland), &
                  land%soil_water        (nzg,iland), &
                  land%soil_energy       (nzg,iland), &
                  head                   (nzg,iland), &
                  land%specifheat_drysoil(nzg,iland)  )

     !-----------------------------------------------------------------------------
     ! TEMPORARY UNTIL FULL LEAF-HYDRO MODEL IS COMPLETED WITH STREAM/RIVER RUNOFF:
     ! Simple representation of runoff

     !s call remove_runoff(iland, iwsfc,     nlev_sfcwater,    &
     !s                    sfcwater_fracliq, sfcwater_mass,    &
     !s                    sfcwater_tempk,   sfcwater_energy,  &
     !s                    sfcwater_depth,   runoff,           &
     !s                    qrunoff                             )
     !-----------------------------------------------------------------------------

     ! Call land patch plot routine for selected iland values.


!  if (iland == 129192) print*, 'swat19 ',land%soil_water(nzg,iland),land%sfcwater_mass(1,iland)

     if (iland == iland_print) call leaf_plot(                &
        iland,                                                &
        land%nlev_sfcwater(iland),                            &
        linit            = 1,                                 &
        lframe           = 1,                                 &
        ktrans           = ktrans,                            &
        sfcwater_tempk   = sfcwater_tempk      (1:nzs),       &
        sfcwater_fracliq = sfcwater_fracliq    (1:nzs),       &
        soil_tempk       = soil_tempk          (1:nzg),       &
        soil_fracliq     = soil_fracliq        (1:nzg),       &
        leaf_class       = sfcg%leaf_class           (iwsfc), &
        can_depth        = sfcg%can_depth            (iwsfc), &
        rhos             = sfcg%rhos                 (iwsfc), &
        vels             = sfcg%vels                 (iwsfc), &
        prss             = sfcg%prss                 (iwsfc), &
        pcpg             = sfcg%pcpg                 (iwsfc), &
        qpcpg            = sfcg%qpcpg                (iwsfc), &
        dpcpg            = sfcg%dpcpg                (iwsfc), &
        sxfer_t          = sfcg%sxfer_t              (iwsfc), &
        sxfer_r          = sfcg%sxfer_r              (iwsfc), &
        ustar            = sfcg%ustar                (iwsfc), &
        cantemp          = sfcg%cantemp              (iwsfc), &
        canshv           = sfcg%canshv               (iwsfc), &
        rshort           = sfcg%rshort               (iwsfc), &
        head1            = sfcg%head1                (iwsfc), &
        head0            = land%head0                (iland), &
        rshort_v         = land%rshort_v             (iland), &
        rshort_g         = land%rshort_g             (iland), &
        rlong_v          = land%rlong_v              (iland), &
        rlong_s          = land%rlong_s              (iland), &
        rlong_g          = land%rlong_g              (iland), &
        veg_height       = land%veg_height           (iland), &
        veg_rough        = land%veg_rough            (iland), &
        veg_tai          = land%veg_tai              (iland), &
        veg_lai          = land%veg_lai              (iland), &
        hcapveg          = land%hcapveg              (iland), &
        snowfac          = land%snowfac              (iland), &
        vf               = land%vf                   (iland), &
        surface_ssh      = land%surface_ssh          (iland), &
        ground_shv       = land%ground_shv           (iland), &
        veg_water        = land%veg_water            (iland), &
        veg_temp         = land%veg_temp             (iland), &
        transp           = transp                           , &
        stom_resist      = land%stom_resist          (iland), &
        veg_fracarea     = land%veg_fracarea         (iland), &
        sfcwater_mass    = land%sfcwater_mass  (1:nzs,iland), &
        sfcwater_energy  = land%sfcwater_energy(1:nzs,iland), &
        sfcwater_depth   = land%sfcwater_depth (1:nzs,iland), &
        soil_water       = land%soil_water     (1:nzg,iland), &
        soil_energy      = land%soil_energy    (1:nzg,iland), &
        head             = head                (1:nzg,iland)  )

     ! At this point, any precipitation has been added to vegetation and surface
     ! water, so zero out precipitation variables.

     sfcg%pcpg (iwsfc) = 0.
     sfcg%qpcpg(iwsfc) = 0.
     sfcg%dpcpg(iwsfc) = 0.



!  if (iland == 129192) print*, 'swat20 ',land%soil_water(nzg,iland),land%sfcwater_mass(1,iland)

     elseif(land%ed_flag(iland) == 1)then

#ifdef USE_ED2
      ! Time to call ED.
      call ed_biophys_wrapper()
#endif

     endif

     ! Zero out SFCG%SXFER_T(iwsfc) and SFCG%SXFER_R(iwsfc) now that they have 
     ! been applied to the canopy

      sfcg%sxfer_t(iwsfc) = 0.
      sfcg%sxfer_r(iwsfc) = 0.
     !sfcg%sxfer_c(iwsfc) = 0.


!  if (iland == 129192) print*, 'swat21 ',land%soil_water(nzg,iland),land%sfcwater_mass(1,iland)

 
!if (iland == 29298) print*, 'le7 ',land%soil_water(nzg,iland),dheight(nzg)

  enddo
  !$omp end do
  !$omp end parallel

end subroutine leaf4

!===============================================================================

subroutine head_column(nzg, iland, slzt, &
                       soil_water, wresid_vg, wsat_vg, alpha_vg, en_vg, &
                       head_press, head, head_slope, soil_watfrac)

  ! Diagnose hydraulic head (relative to local topographic datum) and
  ! its derivative with respect to soil water, and perform any necessary
  ! temporal relaxation of head_press.

  use leaf_coms,  only: dt_leaf
  use leaf4_soil, only: soil_wat2pot

  implicit none

  integer, intent(in) :: nzg, iland

  real, intent(in)    :: slzt        (nzg)
  real, intent(in)    :: soil_water  (nzg)
  real, intent(in)    :: wresid_vg   (nzg)
  real, intent(in)    :: wsat_vg     (nzg)
  real, intent(in)    :: alpha_vg    (nzg)
  real, intent(in)    :: en_vg       (nzg)
  real, intent(inout) :: head_press  (nzg)
  real, intent(inout) :: head        (nzg)
  real, intent(inout) :: head_slope  (nzg)
  real, intent(inout) :: soil_watfrac(nzg)

  real, parameter :: hptimi = 1.0 / 900.       ! Inverse time scale for head_press adjustment [1/s]
  real, parameter :: head_slope_trans = 1000.  ! Inverse specific storage (for transients)
                                               ! of single grid cell at supersaturation

  integer :: k

  real :: psi, psi_slope
  real :: head_press0
  real :: soil_watfrac_ul
  real :: head_press_change

  do k = 1,nzg

     soil_watfrac_ul = (soil_water(k) - wresid_vg(k)) / (wsat_vg(k) - wresid_vg(k))

     soil_watfrac(k) = min(1.0,max(0.001,soil_watfrac_ul))

     call soil_wat2pot(k, iland, soil_water(k), wresid_vg(k), wsat_vg(k), &
                       alpha_vg(k), en_vg(k), psi, psi_slope)

!if (iland == 29298 .and. k == nzg) then
!   print*, 'he1 ',soil_water(k),soil_watfrac_ul,soil_watfrac(k)
!   print*, 'he2 ',alpha_vg(k),wresid_vg(k,iland),wsat_vg(k),slzt(k)
!   print*, 'he3 ',en_vg(k),psi, psi_slope, head(k)
!   print*, 'he4 ',head_press(k)
!endif

     ! Check if pressure head is required and/or active

     if (psi > 1.e-2 .or. head_press(k) > 1.e-2) then

        ! Soil water potential exceeds zero and/or it did recently resulting in
        ! a nonzero value of land%head_press.  Adjust head_press(k) toward psi
        ! at rate defined by relaxation time scale.

        head_press_change = dt_leaf * hptimi * (psi - head_press(k))

        head_press(k) = max(0.,head_press(k) + head_press_change)
           
        ! Compute total head relative to local topographic datum

        head(k) = head_press(k) + slzt(k)
        head_slope(k) = head_slope_trans

     else

        head_press(k) = 0.
        head(k) = psi + slzt(k)
        head_slope(k) = psi_slope

     endif

!if (iland == 29298 .and. k == nzg) then
!   print*, 'he5 ',head_slope(k),head_press0,head_press_change,head(k)
!endif

  enddo

end subroutine head_column

!===============================================================================

subroutine grndvap(iland, rhos, canshv, nlev_sfcwater, surface_ssh, &
                   ground_shv, sfcwater_energy, soil_water, soil_energy, &
                   head, specifheat_drysoil)

  use mem_land,    only: nzg
  use consts_coms, only: grav, rvap
  use misc_coms,   only: io6
  use therm_lib,   only: qtk, rhovsil, qwtk

  implicit none

  integer, intent(in)  :: iland           ! current land cell number 
  real,    intent(in)  :: rhos            ! air density [kg/m^3]
  real,    intent(in)  :: canshv          ! canopy vapor spec hum [kg_vap/kg_air]
  integer, intent(in)  :: nlev_sfcwater   ! # active levels of surface water
  real,    intent(out) :: surface_ssh     ! surface (saturation) spec hum [kg_vap/kg_air]
  real,    intent(out) :: ground_shv      ! ground equilibrium spec hum [kg_vap/kg_air]
  real,    intent(in)  :: sfcwater_energy ! [J/kg]
  real,    intent(in)  :: soil_water      ! soil water content [vol_water/vol_tot]
  real,    intent(in)  :: soil_energy     ! [J/m^3]
  real,    intent(in)  :: head            ! hydraulic head of top soil layer [m]
  real,    intent(in)  :: specifheat_drysoil

  ! Local variables

  real :: tempk     ! surface water temp [K]
  real :: fracliq   ! fraction of surface water in liquid phase
  real :: can_rhov  ! canopy water vapor density [kg_vap/m^3]
  real :: sfc_rhovs ! ground sfc saturation vapor density [kg_vap/m^3]
  real :: gnd_rhov  ! ground sfc evaporative vapor density [kg_vap/m^3]

  if (nlev_sfcwater > 0) then

     ! With surface water (or snowcover) present, sfc_rhovs is the saturation
     ! vapor density of the water/snow surface and is used for computing
     ! dew/frost formation and sfcwater evaporation.

     call qtk(sfcwater_energy,tempk,fracliq)
     sfc_rhovs = rhovsil(tempk-273.15)
     surface_ssh = sfc_rhovs / rhos

  else

     ! Without snowcover, sfc_rhovs is the saturation vapor density at the
     ! temperature of the soil surface, and is used for computing dew/frost
     ! formation only.  

     call qwtk(soil_energy,soil_water*1.e3,specifheat_drysoil,tempk,fracliq)
     sfc_rhovs = rhovsil(tempk-273.15)

     ! gnd_rhov is the effective saturation vapor density of soil and is used
     ! for computing soil evaporation.

     can_rhov = canshv * rhos

     call grndvap_ab(iland,tempk,soil_water,head,can_rhov,sfc_rhovs,gnd_rhov)

     surface_ssh = sfc_rhovs / rhos
     ground_shv  = gnd_rhov  / rhos

  endif

end subroutine grndvap

!===============================================================================

subroutine grndvap_ab(iland,tempk,soil_water,head,can_rhov,sfc_rhovs,gnd_rhov)

  use mem_land,  only: nzg
  use consts_coms, only: grav, rvap
  use misc_coms,   only: io6

  implicit none

  integer, intent(in) :: iland       ! current land cell number 

  real, intent(in)  :: tempk       ! soil temperature [K]
  real, intent(in)  :: soil_water  ! soil water content [vol_water/vol_tot]
  real, intent(in)  :: head        ! hydraulic head [m]
  real, intent(in)  :: can_rhov    ! canopy vapor density [kg_vap/m^3]
  real, intent(in)  :: sfc_rhovs   ! surface saturation vapor density [kg_vap/m^3]
  real, intent(out) :: gnd_rhov    ! ground equilibrium vapor density [kg_vap/m^3]

 ! Local parameter

 real, parameter :: gorvap = grav / rvap  ! gravity divided by vapor gas constant

 ! Local variables

  real :: alpha   ! "alpha" term in Lee and Pielke (1993)
  real :: beta    ! "beta" term in Lee and Pielke (1993)
  real, save :: sfldcap(12)  ! soil water field capacity [vol_water/vol_tot]

!HARDWIRE NTS NOW, BUT LATER COMPUTE SFLDCAP
  integer, parameter :: nts = 5
  data sfldcap/.135,.150,.195,.255,.240,.255,.322,.325,.310,.370,.367,.535/

  ! Without snowcover, gnd_rhov is the effective saturation mixing
  ! ratio of soil and is used for soil evaporation.  First, compute the
  ! "alpha" term or soil "relative humidity" and the "beta" term.
  ! Use limiter on head so that positive values are not used.

  alpha = exp(gorvap * min(0.,head) / tempk)
  beta = .25 * (1. - cos (min(1.,soil_water / sfldcap(nts)) * 3.14159)) ** 2

  gnd_rhov = sfc_rhovs * alpha * beta + (1. - beta) * can_rhov

  ! Require that gnd_rhov be at least slightly less than sfc_rhovs

  gnd_rhov = min(gnd_rhov, 0.999 * sfc_rhovs)

end subroutine grndvap_ab

!===============================================================================

subroutine sfcrad_land(iland, leaf_class, xewl, yewl, zewl, wnxl, wnyl, wnzl, &
                       rshort, rlong, rlongup, rlong_albedo, albedo_beam,     &
                       sfcwater_energy, sfcwater_depth, rshort_s,             &
                       rshort_g, rshort_v, rlong_g, rlong_s, rlong_v,         &
                       nlev_sfcwater, veg_temp, veg_fracarea, veg_height,     &
                       veg_albedo, snowfac, vf, cosz,                         &
                       soil_energy, soil_water, wresid_vg, wsat_vg,           &
                       specifheat_drysoil, sand                               )

  use leaf_coms,   only: nzs, emisv, emisg, emisw
  use consts_coms, only: stefan, eradi
  use misc_coms,   only: io6
  use mem_radiate, only: sunx, suny, sunz
  use therm_lib,   only: qwtk, qtk

  implicit none

  integer, intent(in)  :: iland         ! index of current land cell
  integer, intent(in)  :: leaf_class    ! leaf class
  real,    intent(in)  :: xewl          ! land cell earth x coordinate [m]
  real,    intent(in)  :: yewl          ! land cell earth y coordinate [m]
  real,    intent(in)  :: zewl          ! land cell earth z coordinate [m]
  real,    intent(in)  :: wnxl          ! land cell norm unit vec x comp [m]
  real,    intent(in)  :: wnyl          ! land cell norm unit vec y comp [m]
  real,    intent(in)  :: wnzl          ! land cell norm unit vec z comp [m]
  real,    intent(in)  :: rshort        ! downward surface incident s/w rad flux [W/m^2]
  real,    intent(in)  :: rlong         ! downward surface incident l/w rad flux [W/m^2]
  real,    intent(out) :: rlongup       ! upward sfc l/w rad flux [W/m^2]
  real,    intent(out) :: rlong_albedo  ! albedo for upward sfc l/w
  real,    intent(out) :: albedo_beam   ! land cell albedo (beam)
  real,    intent(in)  :: sfcwater_energy(nzs) ! surface water internal energy [J/kg]
  real,    intent(in)  :: sfcwater_depth(nzs)  ! surface water depth [m]
  real,    intent(out) :: rshort_s(nzs) ! s/w net rad flux to sfc water [W/m^2]
  real,    intent(out) :: rshort_g      ! s/w net rad flux to soil [W/m^2]
  real,    intent(out) :: rshort_v      ! s/w net rad flux to veg [W/m^2]
  real,    intent(out) :: rlong_g       ! l/w net rad flux to soil [W/m^2]
  real,    intent(out) :: rlong_s       ! l/w net rad flux to sfc water [W/m^2]
  real,    intent(out) :: rlong_v       ! l/w net rad flux to veg [W/m^2]
  integer, intent(in)  :: nlev_sfcwater ! # of active surface water layers
  real,    intent(in)  :: veg_temp      ! veg temp [K]
  real,    intent(in)  :: veg_fracarea  ! veg fractional area coverage
  real,    intent(in)  :: veg_height    ! veg height [m]
  real,    intent(in)  :: veg_albedo    ! veg albedo
  real,    intent(out) :: snowfac       ! fractional veg burial by snowcover
  real,    intent(out) :: vf            ! fractional coverage of non-buried part of veg
  real,    intent(out) :: cosz          ! solar zenith angle for land cell
  real,    intent(in)  :: soil_energy   ! soil internal energy [J/m^3]
  real,    intent(in)  :: soil_water    ! soil water content [vol_water/vol_tot]
  real,    intent(in)  :: wresid_vg
  real,    intent(in)  :: wsat_vg
  real,    intent(in)  :: specifheat_drysoil
  real,    intent(in)  :: sand

  ! Local variables

  integer :: k             ! vertical index over surface water layers

  real :: soil_tempk       ! soil temp [K]
  real :: soil_fracliq     ! fraction of soil water in liquid phase
  real :: sfcwater_tempk   ! surface water temp [K]
  real :: sfcwater_fracliq ! fraction of surface water in liquid phase
  real :: rlonga_v         ! longwave radiative flux from atm  to veg  [W/m^2]
  real :: rlonga_s         ! longwave radiative flux from atm  to snow [W/m^2]
  real :: rlonga_g         ! longwave radiative flux from atm  to soil [W/m^2]
  real :: rlongv_a         ! longwave radiative flux from veg  to atm  [W/m^2]
  real :: rlongv_s         ! longwave radiative flux from veg  to snow [W/m^2]
  real :: rlongv_g         ! longwave radiative flux from veg  to soil [W/m^2]
  real :: rlongs_a         ! longwave radiative flux from snow to atm  [W/m^2]
  real :: rlongs_v         ! longwave radiative flux from snow to veg  [W/m^2]
  real :: rlongg_a         ! longwave radiative flux from soil to atm  [W/m^2]
  real :: rlongg_v         ! longwave radiative flux from soil to veg  [W/m^2]
  real :: fracabs          ! fraction of rshort that is absorbed by snowcover
  real :: soil_watfrac_ul  ! soil water fraction (unlimited)
  real :: soil_watfrac     ! soil water fraction (limited)
  real :: vfc       ! 1 - vf
  real :: fc50      ! minimum of soil water fraction and 50%
  real :: alg_dry   ! ground (soil) albedo for dry soil
  real :: alg       ! ground (soil) albedo
  real :: als       ! snowcover albedo (needs better formula based on age of snow)
  real :: alv       ! veg albedo
  real :: abss      ! fraction s/w rad absorbed into top sfcwater layer
  real :: fractrans ! fraction of s/w rad flux transmitted through snowcover layer
  real :: absg      ! fraction of rshort that is absorbed by ground (soil)
  real :: emv       ! veg emissivity
  real :: glong     ! soil l/w rad emission [W/m^2]
  real :: slong     ! sfc water l/w rad emission [W/m^2]
  real :: vlong     ! veg l/w rad emission [W/m^2]

  ! This routine is called twice (for each land cell) by the radiation
  ! parameterization driver, performing exactly the same operations on both calls.
  ! All that is used from the first call are net surface albedo and upward
  ! longwave radiative flux for each land cell.  The radiation parameterization
  ! carries out atmospheric radiative transfer computations following this first 
  ! call using the albedo and upward longwave flux.  The second call to this 
  ! subroutine is made after the atmospheric radiative fluxes are computed.  
  ! This call provides net radiative fluxes to vegetation, snowcover, and soil in
  ! each land cell, plus functions of snowcover, all of which are used in leaf4.

  !-------------------------------------------------------------------------------
  ! Two forms - but still need to consider shadowing and exact energy conservation.
  ! A correct formulation must exchange radiative energy between model columns.

  ! Compute solar zenith angle for land cells

  cosz = (xewl * sunx + yewl * suny + zewl * sunz) * eradi

  ! Compute solar incidence angle for land cells (accounts for local topo slope)

  cosz = wnxl * sunx + wnyl * suny + wnzl * sunz

  ! COSZ IS NOT CURRENTLY USED IN THIS SUBROUTINE
  !-------------------------------------------------------------------------------
  ! Change for LEAF4: rshort_s(1:nzs) is always set to zero here, and rshort_g
  ! always gets all shortwave that is absorbed by the surface, whether it is
  ! sfcwater, soil, or a combination.  Subroutine leaf4_canopy further sorts out
  ! where rshort_g should go.

  rshort_s(1:nzs) = 0.

  if (nlev_sfcwater == 0) then

     ! Case with no surface water

     ! Diagnose soil temperature and liquid fraction

     call qwtk(soil_energy,soil_water*1.e3,  &
               specifheat_drysoil,soil_tempk,soil_fracliq)

     ! Shortwave radiation calculations

     soil_watfrac_ul = (soil_water - wresid_vg) / (wsat_vg - wresid_vg)

     soil_watfrac = min(1.0,max(0.001,soil_watfrac_ul))

     fc50 = min(.50, soil_watfrac)

     if (leaf_class == 2) then

        alg = .80  ! Firn/glacier albedo

     else

        if (sand > 0.7) then
           alg_dry = .41  ! Higher albedo for sandy soil
        else
           alg_dry = .31  ! Soils other than sand and loamy sand
        endif

        alg = alg_dry - .34 * fc50  ! ground (soil) albedo

     endif

     snowfac = 0.

     alv = veg_albedo

     vf = veg_fracarea
     vfc = 1. - vf

     albedo_beam = vf * alv + vfc * vfc * alg

     rshort_g = rshort * vfc * (1. - alg)
     rshort_v = rshort * vf * (1. - alv + vfc * alg)
    !rshort_a = rshort * albedo_beam

     ! Longwave radiation calculations

     emv   = emisv(leaf_class)
     glong = emisg * stefan * soil_tempk ** 4
     vlong = emv * stefan * veg_temp ** 4

     rlonga_v = rlong * vf * (emv + vfc * (1. - emisg))
     rlonga_g = rlong * vfc * emisg
     rlongv_g = vlong * vf * emisg
     rlongv_a = vlong * vf * (2. - emisg - vf + emisg * vf)
     rlongg_v = glong * vf * emv
     rlongg_a = glong * vfc

     rlong_albedo = (vf * (1. - emv) + vfc * vfc * (1. - emisg))

     rlongup = rlongv_a + rlongg_a

     rlong_g = rlonga_g - rlongg_a + rlongv_g - rlongg_v
     rlong_s = 0.
     rlong_v = rlonga_v - rlongv_a + rlongg_v - rlongv_g

  else

     ! Case with surface water

     ! Diagnose surface water temperature and liquid fraction

     call qtk(sfcwater_energy(nlev_sfcwater),sfcwater_tempk,sfcwater_fracliq)

     ! Shortwave radiation calculations

     if (leaf_class == 2) then

        ! Sfcwater albedo ALS is set to .80 over firn/glacier

        als = .80  ! Firn/glacier albedo

     else

        ! Sfcwater albedo ALS ranges from wet-soil value .14 for all-liquid
        ! to .5 for all-ice

        als = .5 - .36 * sfcwater_fracliq

     endif

     ! Sum over sfcwater layers to get total depth

     snowfac = 0.

     do k = 1,nlev_sfcwater
        snowfac = snowfac + sfcwater_depth(k)
     enddo

     snowfac = snowfac / max(.001,veg_height)

     ! If sfcwater (snowcover) is deep enough to cover most vegetation, assume that 
     ! all is covered

     if (snowfac > .9) snowfac = 1.   

     alv = veg_albedo

     vf = veg_fracarea * (1. - snowfac)
     vfc = 1. - vf

     albedo_beam = vf * alv + vfc * vfc * als

     rshort_g = rshort * vfc * (1. - als)
     rshort_v = rshort * vf * (1. - alv + vfc * als)
    !rshort_a = rshort * albedo_beam

     ! Longwave radiation calculations

     emv   = emisv(leaf_class)
     slong = emisw * stefan * sfcwater_tempk ** 4
     vlong = emv * stefan * veg_temp ** 4

     rlonga_v = rlong * vf * (emv + vfc * (1. - emisw))
     rlonga_s = rlong * vfc * emisw
     rlongv_s = vlong * vf * emisw
     rlongv_a = vlong * vf * (2. - emisw - vf + emisw * vf)
     rlongs_v = slong * vf * emv
     rlongs_a = slong * vfc

     rlong_albedo = (vf * (1. - emv) + vfc * vfc * (1. - emisw))

     rlongup = rlongv_a + rlongs_a

     rlong_g = 0.
     rlong_s = rlonga_s - rlongs_a + rlongv_s - rlongs_v
     rlong_v = rlonga_v - rlongv_a + rlongs_v - rlongv_s

  endif

end subroutine sfcrad_land

