!===============================================================================
! OLAM was originally developed at Duke University by Robert Walko, Martin Otte,
! and David Medvigy in the project group headed by Roni Avissar.  Development
! has continued by the same team working at other institutions (University of
! Miami (rwalko@rsmas.miami.edu), the Environmental Protection Agency, and
! Princeton University), with significant contributions from other people.

! Portions of this software are copied or derived from the RAMS software
! package.  The following copyright notice pertains to RAMS and its derivatives,
! including OLAM:  

   !----------------------------------------------------------------------------
   ! Copyright (C) 1991-2006  ; All Rights Reserved ; Colorado State University; 
   ! Colorado State University Research Foundation ; ATMET, LLC 

   ! This software is free software; you can redistribute it and/or modify it 
   ! under the terms of the GNU General Public License as published by the Free
   ! Software Foundation; either version 2 of the License, or (at your option)
   ! any later version. 

   ! This software is distributed in the hope that it will be useful, but
   ! WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   ! or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   ! for more details.
 
   ! You should have received a copy of the GNU General Public License along
   ! with this program; if not, write to the Free Software Foundation, Inc.,
   ! 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA 
   ! (http://www.gnu.org/licenses/gpl.html) 
   !----------------------------------------------------------------------------

!===============================================================================
subroutine spring_dynamics(mrows, moveint, ngr, nma, nua, nwa, xem, yem, zem, &
                           itab_md, itab_ud, itab_wd)

  use mem_ijtabs,  only: itab_md_vars, itab_ud_vars, itab_wd_vars
  use mem_grid,    only: impent
  use consts_coms, only: pi2, erad, erador5, r8
  use misc_coms,   only: io6, nxp, mdomain, deltax
  use oplot_coms,  only: op

  implicit none

  integer, intent(in) :: mrows, moveint, ngr, nma, nua, nwa

  real, intent(inout) :: xem(nma), yem(nma), zem(nma)

  type (itab_md_vars), intent(inout) :: itab_md(nma)
  type (itab_ud_vars), intent(inout) :: itab_ud(nua)
  type (itab_wd_vars), intent(inout) :: itab_wd(nwa)

  integer :: niter
  real, parameter :: relax = .04, beta = 1.2

  ! Automatic arrays

  real :: dist (nua)
  real :: dist0(nua)

  real :: dx(nua)
  real :: dy(nua)
  real :: dz(nua)

  real :: dirs(nma,7)

  integer :: iu,iu1,iu2,iu3,iu4
  integer :: im,im1,im2,im3,im4
  integer :: iw,iw1,iw2,j
  integer :: iter,ipent,mrow1,mrow2,npoly,ngrw,mrmax,mrmin,nmovem

  integer :: iskip
  real :: xp1,xp2,yp1,yp2,xq1,xq2,yq1,yq2

  real :: cosphi3, cosphi4, ratio

  real :: dist00,distm,expansion,frac_change
  character(10) :: string

  integer :: ipic

  ! Array MOVEM is used to flag selected MD points that will be allowed to move
  ! in the spring dynamics procedure.  If MOVEM = 1 the point is allowed to move,
  ! and if MOVEM = 0 the point remains stationary.

  ! Starting with OLAM version 5.2.0, MD points OUTSIDE the current newly refined
  ! mesh are NEVER allowed to move.  This is because ATM and SURFACE grids can be
  ! independently refined, and neither must be allowed to impact the coarser grids
  ! that are common to both ATM and SURFACE.

  integer :: movem(nma)

  real(r8) :: xem8(nma),yem8(nma),zem8(nma)

  logical, parameter :: do_interior = .true.

  xem8(:) = real(xem(:),r8)
  yem8(:) = real(yem(:),r8)
  zem8(:) = real(zem(:),r8)

  ! Use same value of niter for all NGR values in case new grids are added in
  ! subsequent HISTADDGRID run.

  niter = 1000
             
  ! First, turn off movement of ALL M points

  movem(:) = 0

  ! Loop over all M points

  do im = 2,nma

     ! Check grid number of M point.  M point will move in spring dynamics only
     ! if it is of the current grid number (ngr) that was just spawned.  On the
     ! atmosphere grid, the row containing 7-way vertices (future centers of
     ! heptagon Voronoi cells) is the outermost row of M points that are
     ! designated to be on ngr.  On the surface grid, however, the 7-way
     ! vertices are NOT on ngr but rather on the lower-numbered parent grid.
     ! This prevents these M points from moving in spring dynamics.

     if (itab_md(im)%ngr /= ngr) cycle

     ! If interior M points are flagged to move, then all M points on ngr will
     ! be moved, so set movem flag to 1.

     if (moveint == 1) then

        movem(im) = 1

     else 

        ! Even if interior points are not to be moved, M points in the
        ! transition row must always be moveable, so set their movem flag to 1. 

        npoly = itab_md(im)%npoly

        do j = 1,npoly
           iw = itab_md(im)%iw(j)

           ngrw = itab_wd(iw)%ngr

           if (ngrw /= ngr) cycle

           if (itab_wd(iw)%mrow == -3 .or. &
               itab_wd(iw)%mrow == -2 .or. &
               itab_wd(iw)%mrow == -1 .or. &
               itab_wd(iw)%mrow ==  1) then
              movem(im) = 1
              cycle ! once this point has been flagged, we can move on
           endif

        enddo

     endif

  enddo      

  ! Compute mean length of coarse mesh U segments

  if (mdomain < 2) then
     dist00 = beta * pi2 * erad / (5. * real(nxp))
  else
     dist00 = deltax * sqrt(2.0) / sqrt(sqrt(3.0))
  endif

  ! Compute target length of each triangle U segment
  ! Loop over all U points

  !$omp parallel
  !$omp do private (im1,im2,iw1,iw2,mrow1,mrow2,mrmax,mrmin)
  do iu = 2, nua

     im1 = itab_ud(iu)%im(1)
     im2 = itab_ud(iu)%im(2)

     if (movem(im1) == 0 .and. movem(im2) == 0) cycle

     ! Compute target distance for any MRL value

     dist0(iu) = dist00 * 2.0 ** (1 - itab_ud(iu)%mrlu)

     ! Modified distance in MRL border zone

     if (ngr > 1) then

        iw1 = itab_ud(iu)%iw(1)
        iw2 = itab_ud(iu)%iw(2)

        mrow1 = itab_wd(iw1)%mrow
        mrow2 = itab_wd(iw2)%mrow

        mrmax = max(mrow1,mrow2)
        mrmin = min(mrow1,mrow2)
      
        if (mrows == 3) then

           if     (mrmax == -2 .and. mrmin == -2) then
              dist0(iu) = dist0(iu) *  7. / 6.  !* .90
           elseif (mrmax == -1 .and. mrmin == -2) then
              dist0(iu) = dist0(iu) *  8. / 6.  !* .90
           elseif (mrmax == -1 .and. mrmin == -1) then
              dist0(iu) = dist0(iu) *  9. / 6.  !* .90
           elseif (mrmax == 1 .and. mrmin == -1) then
              dist0(iu) = dist0(iu) * 10. / 6.  !* .90
           elseif (mrmax == 1 .and. mrmin == 1) then
              dist0(iu) = dist0(iu) * 11. / 12. !* .90
           endif

        else ! (mrows = 1)

           if (mrmax == 1 .and. mrmin == 1) then
              dist0(iu) = dist0(iu) * 0.75
           endif

        endif

     endif ! ngr

  enddo
  !$omp end do

  !$omp do private (npoly,j,iu)
  do im = 2, nma

     if (movem(im) == 0) cycle

     npoly = itab_md(im)%npoly
     do j = 1, npoly

        iu = itab_md(im)%iu(j)
        if (itab_ud(iu)%im(2) == im) then
           dirs(im,j) =  relax
        else
           dirs(im,j) = -relax
        endif

     enddo
  enddo
  !$omp end do
  !$omp end parallel

  nmovem = 0
  do im = 2, nma
     if (movem(im) == 0) cycle
     nmovem = nmovem + 1
  enddo
  write(io6,'(a,4i9)') "In spring dynamics: ngr,nma,nmovem,niter = ",ngr,nma,nmovem,niter

  ! Main iteration loop 
  do iter = 1, niter

     ! Compute length of each U segment once per iteration
     ! Loop over all U points

     !$omp parallel
     !$omp do private (iu1,iu3,im1,im2,im3,im4)
     do iu = 2, nua

        iu1 = itab_ud(iu)%iu(1)
        iu3 = itab_ud(iu)%iu(3)

        im1 = itab_ud(iu)%im(1)
        im2 = itab_ud(iu)%im(2)

        if (itab_ud(iu1)%im(1) == im1) then
           im3 = itab_ud(iu1)%im(2)
        else
           im3 = itab_ud(iu1)%im(1)
        endif

        if (itab_ud(iu3)%im(1) == im1) then
           im4 = itab_ud(iu3)%im(2)
        else
           im4 = itab_ud(iu3)%im(1)
        endif

        if ( movem(im1) == 0 .and. movem(im2) == 0 .and. &
             movem(im3) == 0 .and. movem(im4) == 0 ) cycle

        dist(iu) = sqrt( (xem8(im1) - xem8(im2)) ** 2 &
                       + (yem8(im1) - yem8(im2)) ** 2 &
                       + (zem8(im1) - zem8(im2)) ** 2 )
     enddo
     !$omp end do
   
     ! Loop over all U points

     !$omp do private (im1,im2,iu1,iu2,iu3,iu4,cosphi3,cosphi4, &
     !$omp             ratio,distm,frac_change)
     do iu = 2, nua

        im1 = itab_ud(iu)%im(1)
        im2 = itab_ud(iu)%im(2)

        if (movem(im1) == 0 .and. movem(im2) == 0) cycle

        ! Adjustment of dist0 based on opposite angles of triangles

        iu1 = itab_ud(iu)%iu(1)
        iu2 = itab_ud(iu)%iu(2)
        iu3 = itab_ud(iu)%iu(3)
        iu4 = itab_ud(iu)%iu(4)

        ! Compute cosine of angles at IM3 and IM4

        cosphi3 = (dist(iu1)**2 + dist(iu2)**2 - dist(iu)**2) / (2. * dist(iu1) * dist(iu2))
        cosphi4 = (dist(iu3)**2 + dist(iu4)**2 - dist(iu)**2) / (2. * dist(iu3) * dist(iu4))

        ! Ratio of smaller cosine to limiting value of cos(72 deg)

        ratio = min(cosphi3,cosphi4) / .309
      
        ! Decrease dist0 if ratio < 1
        ! (NOTE: could just use .309 limit regardless of npoly)

        if (ratio < 1.) then
           distm = dist0(iu) * ratio
        else
           distm = dist0(iu)
        endif

        ! Fractional change to dist that would make it equal dist0

        frac_change = (distm - dist(iu)) / dist(iu)

        ! Compute components of displacement that gives dist0

        dx(iu) = (xem8(im2) - xem8(im1)) * frac_change
        dy(iu) = (yem8(im2) - yem8(im1)) * frac_change
        dz(iu) = (zem8(im2) - zem8(im1)) * frac_change

     enddo
     !$omp end do
   
     ! Loop over all M points

     !$omp do private(npoly,j,iu,expansion)
     do im = 2, nma
      
        if (movem(im) == 0) cycle

        ! For preventing either polar M point from moving:
        !      if (im == impent(1 )) cycle
        !      if (im == impent(12)) cycle

        ! For preventing all pentagonal points from moving:
        !     if (any(im == impent(1:12)) cycle

        ! Apply the displacement components to each M point

        npoly = itab_md(im)%npoly
        do j = 1, npoly
           iu = itab_md(im)%iu(j)
           xem8(im) = xem8(im) + dirs(im,j) * dx(iu)
           yem8(im) = yem8(im) + dirs(im,j) * dy(iu)
           zem8(im) = zem8(im) + dirs(im,j) * dz(iu)
        enddo

        ! Push M point coordinates out to earth radius
      
        if (mdomain < 2) then
           expansion = erad / sqrt(xem8(im) ** 2 + yem8(im) ** 2 + zem8(im) ** 2)
           xem8(im) = xem8(im) * expansion
           yem8(im) = yem8(im) * expansion
           zem8(im) = zem8(im) * expansion
        endif

     enddo
     !$omp end do
     !$omp end parallel

     ! Section for plotting grid at intermediate stages of spring dynamics adjustment

     if (mod(iter,100) == 1) then

        xem(:) = real(xem8(:))
        yem(:) = real(yem8(:))
        zem(:) = real(zem8(:))

        ! Plot grid lines

        call o_reopnwk()

        do ipic = 1,3

        call plotback()
        call oplot_set(ipic)
 
        do iu = 2,nua
           im1 = itab_ud(iu)%im(1)
           im2 = itab_ud(iu)%im(2)

           call oplot_transform(1,xem(im1),yem(im1),zem(im1),xp1,yp1)
           call oplot_transform(1,xem(im2),yem(im2),zem(im2),xp2,yp2)

           call trunc_segment(xp1,xp2,yp1,yp2,xq1,xq2,yq1,yq2,iskip)

           if (iskip == 1) cycle

           call o_frstpt (xq1,yq1)
           call o_vector (xq2,yq2)
        enddo

        if (ipic == 1) then

           ! print mrow values

           do iw = 2, nwa
              im1 = itab_wd(iw)%im(1)
              im2 = itab_wd(iw)%im(2)
              im3 = itab_wd(iw)%im(3) 

              call oplot_transform(1, (xem(im1)+xem(im2)+xem(im3))/3., &
                                      (yem(im1)+yem(im2)+yem(im3))/3., &
                                      (zem(im1)+zem(im2)+zem(im3))/3., &
                                      xp1, yp1                         )

              if ( xp1 < op%xmin .or.  &
                   xp1 > op%xmax .or.  &
                   yp1 < op%ymin .or.  &
                   yp1 > op%ymax ) cycle

              write(string,'(I0)') itab_wd(iw)%mrow
              call o_plchlq (xp1,yp1,trim(adjustl(string)),0.003,0.,0.)
           enddo

        elseif (ipic == 2) then

           ! print ngr values for M points

           do im = 2, nma

              call oplot_transform(1, xem(im), yem(im), zem(im), xp1, yp1)

              if ( xp1 < op%xmin .or.  &
                   xp1 > op%xmax .or.  &
                   yp1 < op%ymin .or.  &
                   yp1 > op%ymax ) cycle

              write(string,'(I0)') itab_md(im)%ngr
              call o_plchlq (xp1,yp1,trim(adjustl(string)),0.003,0.,0.)
           enddo

        elseif (ipic == 3) then

           ! print im values for M points

           do im = 2, nma

              call oplot_transform(1, xem(im), yem(im), zem(im), xp1, yp1)

              if ( xp1 < op%xmin .or.  &
                   xp1 > op%xmax .or.  &
                   yp1 < op%ymin .or.  &
                   yp1 > op%ymax ) cycle

              write(string,'(I0)') im
              call o_plchlq (xp1,yp1,trim(adjustl(string)),0.003,0.,0.)
           enddo

        endif

        call o_frame()

        enddo

        call o_clswk()

     endif ! mod(iter,*)

  enddo ! iter

  xem(:) = real(xem8(:))
  yem(:) = real(yem8(:))
  zem(:) = real(zem8(:))

end subroutine spring_dynamics
