!===============================================================================
! OLAM was originally developed at Duke University by Robert Walko, Martin Otte,
! and David Medvigy in the project group headed by Roni Avissar.  Development
! has continued by the same team working at other institutions (University of
! Miami (rwalko@rsmas.miami.edu), the Environmental Protection Agency, and
! Princeton University), with significant contributions from other people.

! Portions of this software are copied or derived from the RAMS software
! package.  The following copyright notice pertains to RAMS and its derivatives,
! including OLAM:  

   !----------------------------------------------------------------------------
   ! Copyright (C) 1991-2006  ; All Rights Reserved ; Colorado State University; 
   ! Colorado State University Research Foundation ; ATMET, LLC 

   ! This software is free software; you can redistribute it and/or modify it 
   ! under the terms of the GNU General Public License as published by the Free
   ! Software Foundation; either version 2 of the License, or (at your option)
   ! any later version. 

   ! This software is distributed in the hope that it will be useful, but
   ! WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
   ! or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
   ! for more details.
 
   ! You should have received a copy of the GNU General Public License along
   ! with this program; if not, write to the Free Software Foundation, Inc.,
   ! 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA 
   ! (http://www.gnu.org/licenses/gpl.html) 
   !----------------------------------------------------------------------------

!===============================================================================

! TURB and CUPARM flux scheduling:

! 1. Cuparm gives precip RATES
! 2. Stars gives heat and vapor flux RATES
! 3. Fluxes computed once per interval of dtlong = dtlm(1) = dt_leaf = dt_sea,
!    which (as of July 2011) are hardwired to be all the same.
! 4. Fluxes converted to AMOUNTS TRANSFERED based on rate * dtlm(1) 
! 5. Each time leaf runs, it uses at once all it has gotten and zeroes xfer arrays
! 6. When fluxes are done, they are done for all MRL = 1 atm points

! RADIATIVE fluxes only:

! 1. Radiative fluxes transfer RATES only, with no timestep information

!----------------------------------------------------------------------------
! AS OF JULY 2011, SUBROUTINE SURFACE_TURB_FLUXP IS HARDWIRED FOR 
! DT_LEAF = DTLM(1) AND MRL_LEAF = 1, WHICH FOR A LONG TIME HAS BEEN 
! ASSIGNED IN SUBROUTINE MODSCHED AND USED SUCCESSFULLY.
! IF DT_LEAF EVER GETS LARGE ENOUGH TO CAUSE INSTABILITY, THE INSTABILITY
! SHOULD BE CONTROLLED BY INCREASING CAPACITANCE OF THE LEAF COMPONENTS, 
! NOT BY REDUCING TIMESTEP; I.E., WE WANT LEAF TO BE CAPABLE OF RUNNING ON
! AS LONG A TIMESTEP AS THE ATMOSPHERIC MODEL.
!----------------------------------------------------------------------------

subroutine surface_turb_flux(mrl)

  use leaf_coms,   only: isfcl
  use mem_land,    only: mland, omland, land
  use mem_sea,     only: msea, omsea, sea
  use mem_ijtabs,  only: itab_w, itabg_w, jtab_w, jtw_prog, jtw_wstn
  use mem_sfcg,    only: itab_wsfc, sfcg, mwsfc
  use misc_coms,   only: isubdomain, dtlm
  use mem_grid,    only: lsw, lpw, dzt_bot, arw
  use mem_turb,    only: vkm_sfc, sfluxt, sfluxr, sxfer_tk, sxfer_rk, &
                         ustar, wstar, wtv0, pblh, moli
  use mem_basic,   only: press, rho, theta, tair, sh_v, vxe, vye, vze
  use mem_micro,   only: sh_c
  use consts_coms, only: grav, p00, rocp, cpi, eps_virt, vonk, p00i

  implicit none

  integer, intent(in) :: mrl

  integer :: j,iw,isea,iland
  integer :: ks,kw,ka
  integer :: jsfc, jasfc, iwsfc

  real :: exneri
  real :: my_co2
  real :: canexner, canexneri, cantheta, canthetav
  real :: airthetav, ufree
  real :: shflx

  real, parameter :: onethird = 1./3.

  if (isfcl == 0) then

     ! ISFCL = 0 is the no-LEAF option.  Assign surface fluxes here, noting the
     ! following examples.  SHFLX has units of [K m/s kg/m^3].

     ! Default surface flux = 0 W/m^2

     shflx = 0. * cpi

     ! Example with surface flux = 250 W/m^2

     !  shflx = 250. * cpi

     !$omp parallel do private(iw,ks,kw,exneri)
     do j = 1,jtab_w(jtw_prog)%jend(mrl); iw = jtab_w(jtw_prog)%iw(j)

        sfluxt(iw) = 0.
        sfluxr(iw) = 0.
        ustar (iw) = .1  ! Minimum value

        wstar (iw) = 0.
        wtv0  (iw) = 0.

        do ks = 1,lsw(iw)
           kw = ks + lpw(iw) - 1

           vkm_sfc (ks,iw) = 0.
           sxfer_tk(ks,iw) = 0.
           sxfer_rk(ks,iw) = 0.

           exneri = theta(kw,iw) / tair(kw,iw)

           sxfer_tk(ks,iw) = dtlm(1) * shflx * exneri &
                           * (arw(kw,iw) - arw(kw-1,iw))

        enddo

        ! albedt (iw) = albedo
        ! rlongup(iw) = stefan * 288.15 ** 4  ! std msl temp; make decision to not
                                              ! run radiation if not running leaf?
     enddo
     !$omp end parallel do

     return
  endif

  ! ISFCL = 1 is the LEAF option...

  ! Reset to zero the atm values of VKM_SFC, USTAR, SFLUXT, and SFLUXR
  ! for same mrl's that fluxes will be evaluated for this time.  
  ! THUS, VKM_SFC, USTAR, SFLUXT, AND SFLUXR ARE ONLY SUMMED OVER
  ! SPACE, BUT NOT OVER TIME. (ON THE OTHER HAND, SXFER_TK AND SXFER_RK ARE SUMMED
  ! OVER BOTH SPACE AND TIME; THEY ARE RESET TO ZERO IN THILTEND_LONG AND
  ! SCALAR_TRANSPORT AFTER THEY ARE TRANSFERRED TO THE ATMOSPHERE.)

  ! Set sea and land fluxes to be done for SURFACE SIMILARITY: 
  !    Do fluxes at beginning of long timestep at given mrl

  !$omp parallel do private(iw)
  do j = 1,jtab_w(jtw_wstn)%jend(mrl); iw = jtab_w(jtw_wstn)%iw(j)

     vkm_sfc(:,iw) = 0.
     ustar    (iw) = 0.
     wtv0     (iw) = 0.
     sfluxt   (iw) = 0.
     sfluxr   (iw) = 0.

     ! for now, zero out surface transfer arrays at the start
     ! of each long timestep:

     sxfer_tk(:,iw) = 0.
     sxfer_rk(:,iw) = 0.
   ! sxfer_ck(:,iw) = 0.  ! placeholder for CO2

  enddo
  !$omp end parallel do

  ! Loop over all SFC grid cells in subdomain, EVEN THOSE THAT ARE NOT PRIMARY,
  ! so that all surface fluxes are computed beneath all ATM columns that are primary

  !dir$ novector
  !$omp parallel do private(airthetav,canexner,canexneri,cantheta,canthetav, &
  !$omp                     ufree,iland,my_co2,isea)
  do iwsfc = 2,mwsfc

     airthetav = sfcg%airtheta(iwsfc) * (1.0 + eps_virt * sfcg%airshv(iwsfc))
     canexner  = (sfcg%prss(iwsfc) * p00i) ** rocp
     canexneri = 1. / canexner
     cantheta  = sfcg%cantemp(iwsfc) * canexneri
     canthetav = cantheta * (1.0 + eps_virt * sfcg%canshv(iwsfc))
     ufree     = (grav / airthetav * max(sfcg%wthv(iwsfc),0.0) * sfcg%dzt_bot(iwsfc)) ** onethird

     ! Compute turbulent fluxes based on whether SFC grid cell is land/ED2, land, lake, or sea

     if (sfcg%leaf_class(iwsfc) >= 2) then
        if (land%ed_flag(iwsfc-omland) > 0) then
           iland = iwsfc - omland

           ! This is land cell that is flagged for processing by ED2

#ifdef USE_ED2

           ! Someday we may track CO2 in OLAM...
           call get_ed2_atm_co2(iland,my_co2)

           call ed_stars_wrapper(  iland, &
                    sfcg%dzt_bot (iwsfc), &
                    sfcg%vels    (iwsfc), &
                    sfcg%rhos    (iwsfc), &
                    ufree               , &
                    sfcg%airtheta(iwsfc), &
                    airthetav           , &
                    sfcg%airshv  (iwsfc), &
                    my_co2,             , &
                    sfcg%vkmsfc  (iwsfc), &
                    sfcg%sfluxt  (iwsfc), &
                    sfcg%sfluxr  (iwsfc), &
                    sfcg%sfluxc  (iwsfc), &
                    sfcg%ustar   (iwsfc), &
                    sfcg%rib     (iwsfc), &
                    sfcg%ggaer   (iwsfc), &
                    sfcg%wthv    (iwsfc), &
                    sfcg%zeta    (iwsfc)  )

#endif

           sfcg%sxfer_t(iwsfc) = dtlm(1) * sfcg%sfluxt(iwsfc) * canexner
           sfcg%sxfer_r(iwsfc) = dtlm(1) * sfcg%sfluxr(iwsfc)

           ! If we have co2 or other scalars:

           !sfcg%sfluxc(iwsfc) = sfcg%rhos(iwsfc) * sfcg%ggaer(iwsfc) * (sfcg%canco2(iwsfc) - sfcg%airco2(iwsfc))
           !sfcg%sxfer_c(iwsfc) = dtlm(1) * sfcg%sfluxc(iwsfc)

           ! Processing of this ED2 land cell is complete, so cycle to next cell

           cycle
        endif
     endif

     ! Only non-ED2 cells remain at this point in the loop

     if (sfcg%leaf_class(iwsfc) >= 1) then

        ! This is a land or lake cell

        call stars(sfcg%dzt_bot (iwsfc), &
                   sfcg%rough   (iwsfc), &
                   sfcg%vels    (iwsfc), &
                   sfcg%rhos    (iwsfc), &
                   ufree               , &
                   sfcg%airtheta(iwsfc), &
                   airthetav           , &
                   sfcg%airshv  (iwsfc), &
                   cantheta            , &
                   canthetav           , &
                   sfcg%canshv  (iwsfc), &
                   sfcg%vkmsfc  (iwsfc), &
                   sfcg%sfluxt  (iwsfc), &
                   sfcg%sfluxr  (iwsfc), & 
                   sfcg%ustar   (iwsfc), &
                   sfcg%rib     (iwsfc), &
                   sfcg%ggaer   (iwsfc)  )

!if (iwsfc == 4182) then
!   print*, ' '
!   print*, 'sfl1 ',sfcg%sfluxr(iwsfc),sfcg%sfluxt(iwsfc)
!   print*, 'sfl2 ',cantheta,canthetav,ufree
!   print*, 'sfl3 ',sfcg%airtheta(iwsfc),sfcg%airshv(iwsfc),sfcg%prss(iwsfc)
!   print*, 'sfl4 ',sfcg%rough(iwsfc),sfcg%vels(iwsfc),sfcg%rhos(iwsfc)
!   print*, 'sfl5 ',sfcg%dzt_bot(iwsfc),airthetav,sfcg%canshv(iwsfc)
!   print*, 'sfl6 ',sfcg%ustar(iwsfc)
!endif

     else

        ! This is sea cell.  First, compute turbulent fluxes over open water areas.

        isea = iwsfc - omsea

        cantheta  = sea%sea_cantemp(isea) * canexneri
        canthetav = cantheta * (1.0 + eps_virt * sea%sea_canshv(isea))

        ufree = (grav / airthetav * max(sea%sea_wthv(isea),0.0) * sfcg%dzt_bot(iwsfc)) ** onethird

        call stars(sfcg%dzt_bot  (iwsfc), &
                   sea%sea_rough  (isea), &
                   sfcg%vels     (iwsfc), &
                   sfcg%rhos     (iwsfc), &
                   ufree                , &
                   sfcg%airtheta (iwsfc), &
                   airthetav            , &
                   sfcg%airshv   (iwsfc), &
                   cantheta             , &
                   canthetav            , &
                   sea%sea_canshv (isea), &
                   sea%sea_vkmsfc (isea), &
                   sea%sea_sfluxt (isea), &
                   sea%sea_sfluxr (isea), & 
                   sea%sea_ustar  (isea), &
                   sea%sea_rib    (isea), &
                   sea%sea_ggaer  (isea)  )

        sea%sea_wthv(isea) = ( sea%sea_sfluxt(isea) * (1.0 + eps_virt * sfcg%airshv(iwsfc)) &
           + sea%sea_sfluxr(isea) * eps_virt * sfcg%airtheta(iwsfc) ) / sfcg%rhos(iwsfc)

        sea%sea_zeta(isea) = sfcg%dzt_bot(iwsfc) * grav * vonk * sea%sea_wthv(isea)  &
                           / (airthetav * sea%sea_ustar(isea) ** 3)


!if (iwsfc == 4182) then
!   print*, ' '
!   print*, 'sfl11 ',isea,sea%nlev_seaice(isea),sea%sea_sfluxr(isea),sea%sea_sfluxt(isea)
!   print*, 'sfl12 ',cantheta,canthetav,ufree
!   print*, 'sfl13 ',sfcg%airtheta(iwsfc),sfcg%airshv(iwsfc),sfcg%prss(iwsfc)
!endif



        ! When we have CO2:
        ! sea%sea_sfluxc(isea) = rhos * sea_ggaer(isea) * (sea%sea_co2 - air_co2)
        ! sea%sea_sfluxc(isea) = 0.

        ! Check if sea ice is present

        if (sea%nlev_seaice(isea) == 0) then

           ! If no sea ice is present in this cell, zero out fluxes over ice
           ! and set cell flux to open water part

           sfcg%vkmsfc(iwsfc) = sea%sea_vkmsfc(isea)
           sfcg%ustar (iwsfc) = sea%sea_ustar (isea)
           sfcg%ggaer (iwsfc) = sea%sea_ggaer (isea)
           sfcg%sfluxt(iwsfc) = sea%sea_sfluxt(isea)
           sfcg%sfluxr(iwsfc) = sea%sea_sfluxr(isea)
           sfcg%wthv  (iwsfc) = sea%sea_wthv  (isea)
           sfcg%zeta  (iwsfc) = sea%sea_zeta  (isea)
           sfcg%rib   (iwsfc) = sea%sea_rib   (isea)
         ! sfcg%sfluxc(iwsfc) = sea%sea_sfluxc(isea)

           sea%ice_vkmsfc(isea) = 0.0
           sea%ice_ustar (isea) = 0.0
           sea%ice_ggaer (isea) = 0.0
           sea%ice_sfluxt(isea) = 0.0
           sea%ice_sfluxr(isea) = 0.0
         ! sea%ice_sfluxc(isea) = 0.0

        else
      
           ! If sea ice is present in this cell, compute turbulent fluxes over ice

           cantheta  = sea%ice_cantemp(isea) * canexneri
           canthetav = cantheta * (1.0 + eps_virt * sea%ice_canshv(isea))

           ufree = (grav / airthetav * max(sea%ice_wthv(isea),0.0) * sfcg%dzt_bot(iwsfc)) ** onethird

           call stars(sfcg%dzt_bot  (iwsfc), &
                      sea%ice_rough  (isea), &
                      sfcg%vels     (iwsfc), &
                      sfcg%rhos     (iwsfc), &
                      ufree                , &
                      sfcg%airtheta (iwsfc), &
                      airthetav            , &
                      sfcg%airshv   (iwsfc), &
                      cantheta             , &
                      canthetav            , &
                      sea%ice_canshv (isea), &
                      sea%ice_vkmsfc (isea), &
                      sea%ice_sfluxt (isea), &
                      sea%ice_sfluxr (isea), &
                      sea%ice_ustar  (isea), &
                      sea%ice_rib    (isea), &
                      sea%ice_ggaer  (isea)  )

           sea%ice_wthv(isea) = ( sea%ice_sfluxt(isea) * (1.0 + eps_virt * sfcg%airshv(iwsfc)) &
              + sea%ice_sfluxr(isea) * eps_virt * sfcg%airtheta(iwsfc) ) / sfcg%rhos(iwsfc)

           sea%ice_zeta(isea) = sfcg%dzt_bot(iwsfc) * grav * vonk * sea%ice_wthv(isea)  &
                              / (airthetav * sea%ice_ustar(isea) ** 3)

           ! When we have CO2:
           ! sea%ice_sfluxc(isea) = sfcg%rhos(iwsfc) * sea_ggaer(isea) * (sea%ice_co2(isea) - sea%airco2(isea))
           ! sea%ice_sfluxc(isea) = 0.

           ! Combine sea and ice values based on ice fraction:

           sfcg%vkmsfc(iwsfc) = (1.0 - sea%seaicec(isea)) * sea%sea_vkmsfc(isea) &
                                     + sea%seaicec(isea)  * sea%ice_vkmsfc(isea)

           sfcg%ustar(iwsfc)  = (1.0 - sea%seaicec(isea)) * sea%sea_ustar(isea) &
                                     + sea%seaicec(isea)  * sea%ice_ustar(isea) 

           sfcg%ggaer(iwsfc)  = (1.0 - sea%seaicec(isea)) * sea%sea_ggaer(isea) &
                                     + sea%seaicec(isea)  * sea%ice_ggaer(isea) 

           sfcg%sfluxt(iwsfc) = (1.0 - sea%seaicec(isea)) * sea%sea_sfluxt(isea) &
                                     + sea%seaicec(isea)  * sea%ice_sfluxt(isea)

           sfcg%sfluxr(iwsfc) = (1.0 - sea%seaicec(isea)) * sea%sea_sfluxr(isea) &
                                    + sea%seaicec(isea)  * sea%ice_sfluxr(isea)

           sfcg%wthv(iwsfc)   = (1.0 - sea%seaicec(isea)) * sea%sea_wthv(isea) &
                                     + sea%seaicec(isea)  * sea%ice_wthv(isea)

           sfcg%zeta(iwsfc)   = (1.0 - sea%seaicec(isea)) * sea%sea_zeta(isea) &
                                     + sea%seaicec(isea)  * sea%ice_zeta(isea)

           sfcg%rib(iwsfc)    = (1.0 - sea%seaicec(isea)) * sea%sea_rib(isea) &
                                     + sea%seaicec(isea)  * sea%ice_rib(isea)

           ! sfcg%sfluxc(iwsfc) = (1.0 - sea%seaicec(isea)) * sea%sea_sfluxc(isea) &
           !                           + sea%seaicec(isea)  * sea%ice_sfluxc(isea)


if (iwsfc == 4182) then
   print*, ' '
   print*, 'sfl24 ',isea,sea%nlev_seaice(isea),sea%sea_sfluxr(isea),sea%sea_sfluxt(isea)
   print*, 'sfl25 ',cantheta,canthetav,ufree
   print*, 'sfl26 ',sea%seaicec(isea),sea%ice_sfluxt(isea),sea%ice_sfluxr(isea)
   print*, 'sfl27 ',sea%ice_cantemp(isea)
endif

        endif

        ! Store flux contributions in SEA cell

        sea%sea_sxfer_t(isea) = dtlm(1) * sea%sea_sfluxt(isea) * canexner
        sea%ice_sxfer_t(isea) = dtlm(1) * sea%ice_sfluxt(isea) * canexner

        sea%sea_sxfer_r(isea) = dtlm(1) * sea%sea_sfluxr(isea)
        sea%ice_sxfer_r(isea) = dtlm(1) * sea%ice_sfluxr(isea)

      ! sea%sea_sxfer_c(isea) = dtlm(1) * sea%sea_sfluxc(isea)
      ! sea%ice_sxfer_c(isea) = dtlm(1) * sea%ice_sfluxc(isea)

     endif  ! if this is sea cell

     ! For any SFC grid cell

     sfcg%sxfer_t(iwsfc) = dtlm(1) * sfcg%sfluxt(iwsfc) * canexner
     sfcg%sxfer_r(iwsfc) = dtlm(1) * sfcg%sfluxr(iwsfc)

     ! If we have co2 or other scalars:
     ! sfcg%sfluxc(iwsfc) = sfcg%rhos(iwsfc) * sfcg%ggaer(iwsfc) * (sfcg%canco2(iwsfc) - sfcg%airco2(iwsfc))
     ! sfcg%sxfer_c(iwsfc) = dtlm(1) * sfcg%sfluxc(iwsfc)

  enddo
  !$omp end parallel do

  ! Loop over ATM grid columns that are primary in this subdomain

  !$omp parallel do private(iw, jsfc, iwsfc, jasfc, kw, ka, ks)
  do j = 1,jtab_w(jtw_prog)%jend(mrl); iw = jtab_w(jtw_prog)%iw(j)

     ! Loop over all SFC grid cells that couple to this ATM grid column

     do jsfc = 1,itab_w(iw)%jsfc2
        iwsfc = itab_w(iw)%iwsfc(jsfc)
        jasfc = itab_w(iw)%jasfc(jsfc)

        kw = itab_wsfc(iwsfc)%kwatm(jasfc)
        ka = lpw(iw)
        ks = kw - ka + 1

        ! Add flux contributions to IW atmospheric column

        ustar      (iw) = ustar      (iw) + itab_wsfc(iwsfc)%arcoariw(jasfc) * sfcg%ustar (iwsfc) 
        wtv0       (iw) = wtv0       (iw) + itab_wsfc(iwsfc)%arcoariw(jasfc) * sfcg%wthv  (iwsfc)
        sfluxt     (iw) = sfluxt     (iw) + itab_wsfc(iwsfc)%arcoariw(jasfc) * sfcg%sfluxt(iwsfc)
        sfluxr     (iw) = sfluxr     (iw) + itab_wsfc(iwsfc)%arcoariw(jasfc) * sfcg%sfluxr(iwsfc)
        vkm_sfc (ks,iw) = vkm_sfc (ks,iw) + itab_wsfc(iwsfc)%arcoarkw(jasfc) * sfcg%vkmsfc(iwsfc)
                        ! * land%slope_fact(iwsfc-omland)
        sxfer_tk(ks,iw) = sxfer_tk(ks,iw) + itab_wsfc(iwsfc)%arc(jasfc) * dtlm(1) * sfcg%sfluxt(iwsfc)
        sxfer_rk(ks,iw) = sxfer_rk(ks,iw) + itab_wsfc(iwsfc)%arc(jasfc) * dtlm(1) * sfcg%sfluxr(iwsfc)
     enddo
  enddo
  !$omp end parallel do

  ! Compute some derived surface quantities

  !$omp parallel do private(iw,ka)
  do j = 1, jtab_w(jtw_prog)%jend(mrl); iw = jtab_w(jtw_prog)%iw(j)

     ka = lpw(iw)

     moli(iw) = - grav * vonk * wtv0(iw) /  &
                ( ustar(iw)**3 * theta(ka,iw) * (1.0 + eps_virt * sh_v(ka,iw)) )

     if (wtv0(iw) > 0.0) then
        wstar(iw) = (grav * pblh(iw) * wtv0(iw) / theta(ka,iw)) ** onethird
     else
        wstar(iw) = 0.0
     endif

  enddo
  !$omp end parallel do

  ! No MPI send/recv communication of ATM turbulent fluxes are required for 
  ! averaging to SFC grid cells because the fluxes are directly computed on
  ! SFC grid cells.

end subroutine surface_turb_flux

!===============================================================================

subroutine stars( zts, rough, vels, rhos, ufree,  &
                  air_theta, air_thetav, air_shv, &
                  can_theta, can_thetav, can_shv, &
                  vkmsfc, sfluxt, sfluxr, ustar,  &
                  ri, ggaero                      )

  ! Subroutine stars computes surface heat and vapor fluxes and momentum drag
  ! coefficient from Louis (1981) equations

  use consts_coms, only: vonk, grav

  implicit none

  ! Input variables

  real, intent(in) :: zts        ! height above surface of {vels, ths, shv} [m]
  real, intent(in) :: rough      ! surface roughness height [m]
  real, intent(in) :: vels       ! atmos near-surface wind speed [m/s]
  real, intent(in) :: rhos       ! atmos near-surface density [kg/m^3]
  real, intent(in) :: ufree      ! surface layer free-convective velocity [m/s]
  real, intent(in) :: air_theta  ! atmos near-surface pot. temp [K]
  real, intent(in) :: air_thetav ! atmos near-surface virt. pot. temp [K]
  real, intent(in) :: air_shv    ! atmos near-surface vapor spec hum [kg_vap/m^3]
  real, intent(in) :: can_theta  ! canopy air pot. temp [K]
  real, intent(in) :: can_thetav ! canopy air virt. pot. temp [K]
  real, intent(in) :: can_shv    ! canopy air vapor spec hum [kg_vap/m^3]

  ! Output variables

  real, intent(out) :: vkmsfc    ! surface drag coefficient for this flux cell
  real, intent(out) :: sfluxt    ! surface sensible heat flux for this flux cell
  real, intent(out) :: sfluxr    ! surface vapor flux for this flux cell
  real, intent(out) :: ustar     ! surface friction velocity for this flux cell
  real, intent(out) :: ri        ! bulk richardson number, eq. 3.45 in Garratt
  real, intent(out) :: ggaero    ! bare ground conductance m/s

  ! Local parameters

  real, parameter :: b = 5.
  real, parameter :: csm = 7.5
  real, parameter :: csh = 5.
  real, parameter :: d = 5.
  real, parameter :: ustmin = .05 ! lower bound on ustar (friction velocity)
  real, parameter :: ubmin  = .1  ! lower bound on wind speed 

  ! Local variables

  real :: vels0  ! wind speed with minimum imposed [m/s]
  real :: a2     ! drag coefficient in neutral conditions, here same for h/m
  real :: c1
  real :: c2
  real :: c3
  real :: cm
  real :: ch
  real :: fm
  real :: fh
  real :: tstar  ! 
  real :: rstar  !
  real :: vtscr  ! ustar times density

  ! Routine to compute Louis (1981) surface layer parameterization.

  vels0 = max(vels,ubmin,ufree)

  a2 = ( vonk / log(zts / rough) ) ** 2
  c1 = a2 * vels0

  ri = 2.0 * grav * zts * (air_thetav - can_thetav)  &
     / ( (air_thetav + can_thetav) * vels0 * vels0 )

  if (air_thetav >= can_thetav) then

     fm = 1. / (1. + (2. * b * ri / sqrt(1. + d * ri)))
     fh = 1. / (1. + (3. * b * ri * sqrt(1. + d * ri)))

  else                            ! UNSTABLE CASE

     c2 = b * a2 * sqrt(zts / rough * (abs(ri)))
     cm = csm * c2
     ch = csh * c2
     fm = (1. - 2. * b * ri / (1. + 2. * cm))
     fh = (1. - 3. * b * ri / (1. + 3. * ch))

  endif

  ustar = max(ustmin,sqrt(c1 * vels0 * fm))
  c3 = c1 * fh / ustar
  tstar = c3 * (air_theta - can_theta)
  rstar = c3 * (air_shv   - can_shv)

  vtscr = ustar * rhos

  vkmsfc =   vtscr * ustar * zts / vels0
  sfluxt = - vtscr * tstar
  sfluxr = - vtscr * rstar

  ! Store the aerodynamic conductance between the surface canopy and
  ! the lowest model level

  ggaero  = c3 * ustar

end subroutine stars

!===============================================================================

subroutine surface_cuparm_flux()

  use mem_cuparm,  only: conprr
  use mem_ijtabs,  only: istp, mrl_begl, itabg_w
  use mem_sfcg,    only: itab_wsfc
  use consts_coms, only: cliq, alli, t00
  use mem_basic,   only: tair, rho, sh_v
  use leaf_coms,   only: mrl_leaf
  use mem_sfcg,    only: mwsfc, itab_wsfc, sfcg
  use misc_coms,   only: io6, isubdomain, dtlm
  use ed_misc_coms,only: ed2_active
  use therm_lib,   only: rhovsl
  use mem_para,    only: myrank

  implicit none

  integer :: mrl
  integer :: iw
  integer :: iwsfc, j
  integer :: kw

  real :: airtempc
  real :: tempc

  ! Subroutine to transfer atmospheric cumulus parameterization 
  ! precipitation FLUX to surface cells

  ! Set land fluxes to be done for CUPARM
  !    1. for mrl = 0, no fluxes 
  !    2. For mrl > mrl_leaf, no fluxes
  !    3. For mrl <= mrl_leaf, do fluxes at beginning of long timestep 
  !                            for all points in mrl = 1

  mrl = mrl_begl(istp)
  if (mrl > 0) then

     ! Transfer precipitation FLUX to SFC grid cells

     !$omp parallel do private(j, iw, kw, airtempc, tempc) 
     do iwsfc = 2, mwsfc

        ! Skip this cell if running in parallel and cell rank is not MYRANK
        if (isubdomain == 1 .and. itab_wsfc(iwsfc)%irank /= myrank) cycle

        do j = 1,itab_wsfc(iwsfc)%nwatm
           iw = itab_wsfc(iwsfc)%iwatm(j)  ! local index

      !s     ! If run is parallel, convert iw to local domain

      !s     if (isubdomain == 1) then
      !s        iw = itabg_w(iw)%iw_myrank
      !s    endif

           kw = itab_wsfc(iwsfc)%kwatm(j)

           ! Compute air temperature in C

           airtempc = tair(kw,iw) - t00

           ! Estimate wet bulb temp using computation from subroutine each_column in micphys
           ! Assume that convective precip reaches surface at this wet bulb temp

           tempc = airtempc - min(25.,  &
                   700. * (rhovsl(airtempc) / real(rho(kw,iw)) - sh_v(kw,iw)))

           sfcg%pcpg (iwsfc) = sfcg%pcpg (iwsfc) + itab_wsfc(iwsfc)%arcoarsfc(j) * dtlm(1) * conprr(iw)
           sfcg%qpcpg(iwsfc) = sfcg%qpcpg(iwsfc) + itab_wsfc(iwsfc)%arcoarsfc(j) * dtlm(1) * conprr(iw) * (cliq * tempc + alli)
           sfcg%dpcpg(iwsfc) = sfcg%dpcpg(iwsfc) + itab_wsfc(iwsfc)%arcoarsfc(j) * dtlm(1) * conprr(iw) * .001
        enddo

     enddo
     !$omp end parallel do

#ifdef USE_ED2
   if (ed2_active == 1) then
      call copy_cuparm_to_ed()
   endif
#endif

  endif

end subroutine surface_cuparm_flux

