module hdf5_f2f

  use hdf5
  implicit none

  integer(HID_T)   :: fileid
  integer(HID_T)   :: xferid
  integer(HID_T)   :: propid
  integer(HID_T)   :: dsetid
  integer(HID_T)   :: fspcid
  integer(HID_T)   :: mspcid
  integer(HID_T)   :: dspcid
  integer(HSIZE_T) :: dimsf(7)
  integer(HSIZE_T) :: ndimsf

  interface fh5_write
     module procedure                 &
          fh5_write_integer_scalar,   &
          fh5_write_real_scalar,      &
          fh5_write_character_scalar, &
          fh5_write_real8_scalar,     &
          fh5_write_logical_scalar,   &
          fh5_write_integer_array,    &
          fh5_write_real_array,       &
          fh5_write_character_array,  &
          fh5_write_real8_array,      &
          fh5_write_logical_array
  end interface fh5_write

  interface fh5d_read
     module procedure                  &
          fh5d_read_integer_scalar,    &
          fh5d_read_real_scalar,       &
          fh5d_read_character_scalar,  &
          fh5d_read_real8_scalar,      &
          fh5d_read_logical_scalar,    &
          fh5d_read_integer_array,     &
          fh5d_read_real_array,        &
          fh5d_read_character_array,   &
          fh5d_read_real8_array,       &
          fh5d_read_logical_array
  end interface fh5d_read


contains


  subroutine fh5f_open(locfn, iaccess, hdferr)
    implicit none

    character(*), intent(IN)  :: locfn
    integer,      intent(IN)  :: iaccess
    integer,      intent(OUT) :: hdferr

    integer(HID_T) :: access_id
    integer        :: flags
    integer        :: nulo

    ! turn off default error handling
    call h5eset_auto_f(0, nulo)

    access_id = H5P_DEFAULT_F
    if(iaccess == 1) flags = H5F_ACC_RDONLY_F
    if(iaccess == 2) flags = H5F_ACC_RDWR_F

    call h5fopen_f(locfn, flags, fileid, hdferr, access_id)

    hdferr = int(fileid) + hdferr

    return
  end subroutine fh5f_open



  subroutine fh5f_close(hdferr)
    implicit none

    integer, intent(OUT) :: hdferr

    call h5fclose_f(fileid, hdferr)

    return
  end subroutine fh5f_close



  subroutine fh5f_create(locfn, iaccess, hdferr)

#if defined(OLAM_MPI) && defined(OLAM_PARALLEL_HDF5)
    use misc_coms, only: ipar_out, io6
    use mpi
#endif

    implicit none

    character(*), intent(IN)  :: locfn
    integer,      intent(IN)  :: iaccess
    integer,      intent(OUT) :: hdferr

    integer(HID_T) :: access_id
    integer(HID_T) :: create_id
    integer        :: flags
    integer        :: info, ierr

    create_id = H5P_DEFAULT_F
    access_id = H5P_DEFAULT_F
    if (iaccess == 1) flags = H5F_ACC_TRUNC_F
    if (iaccess == 2) flags = H5F_ACC_EXCL_F

#if defined(OLAM_MPI) && defined(OLAM_PARALLEL_HDF5)
    if (ipar_out == 1) then
       write(io6,*)
       write(io6,*) "Enabling parallel HDF5 output"
       write(io6,*)
       
       info = MPI_INFO_NULL

#if defined(__HOS_AIX__) || defined(__TOS_AIX__) || defined(_AIX)
       ! This improved performance with IBM's native GPFS driver
       call MPI_Info_create(info, ierr)
       call MPI_Info_set(info, "IBM_largeblock_io", "true", ierr)
#endif

       call h5pcreate_f(H5P_FILE_ACCESS_F, access_id, hdferr)
       call h5pset_fapl_mpio_f(access_id, MPI_COMM_WORLD, info, hdferr)

#if defined(__HOS_AIX__) || defined(__TOS_AIX__) || defined(_AIX)
       call MPI_Info_free(info, ierr)
#endif

    endif
#endif

    call h5fcreate_f(locfn, flags, fileid, hdferr, create_id, access_id)

#if defined(OLAM_MPI) && defined(OLAM_PARALLEL_HDF5)
    if (ipar_out == 1) then
       call h5pclose_f(access_id, hdferr)
    endif
#endif

    hdferr = int(fileid) + hdferr

    return
  end subroutine fh5f_create



  subroutine fh5_prepare_write(ndims, dims, hdferr, icompress, &
                               mcoords, fcoords, ifsize)

#if defined(OLAM_MPI) && defined(OLAM_PARALLEL_HDF5)
    use misc_coms, only: ipar_out
#endif

    implicit none

    integer, intent(IN)           :: ndims
    integer, intent(IN)           :: dims(:)
    integer, intent(OUT)          :: hdferr
    integer, intent(IN), optional :: icompress
    integer, intent(IN), optional :: mcoords(:), fcoords(:), ifsize

    logical          :: docompress
    integer          :: i, j, iop
    integer(HSIZE_T) :: dims_file(ndims)
    integer(HSIZE_T) :: offset(ndims), countf(ndims)

    ! Output dimensions

    dimsf = 1
    ndimsf = ndims

    do i = 1, ndims
       dimsf(i) = dims(i)
    end do

    dims_file = 1
     do i = 1, ndims
       dims_file(i) = dims(i)
    end do
   
    ! For distributed output, global data (file) size will be different then
    ! the local array size

    if (present(fcoords) .and. present(ifsize)) then
       dims_file(ndims) = ifsize
    endif

    ! Create a property list for compression/chunking/filters

    call h5pcreate_f(H5P_DATASET_CREATE_F, propid, hdferr) 	 

    ! If no parallel IO, activate HDF5 compression for valid icompress

    if (present(icompress)) then

       if (icompress > 0 .and. icompress < 10) then
          docompress = .true.
       else
          docompress = .false.
       endif

#if defined(OLAM_MPI) && defined(OLAM_PARALLEL_HDF5)
       ! Compression does not work with parallel output
       if (ipar_out == 1) docompress = .false.
#endif

       if (docompress) then
          if (.not. (ndims == 1 .and. dimsf(1) == 1)) then
             call h5pset_chunk_f(propid, ndims, dimsf, hdferr)
             call h5pset_shuffle_f(propid, hdferr)
             call h5pset_deflate_f(propid, icompress, hdferr)	 
          endif
       endif
    endif

    ! Create transfer property list for collective dataset write

    xferid = H5P_DEFAULT_F

#if defined(OLAM_MPI) && defined(OLAM_PARALLEL_HDF5)
    if (ipar_out == 1) then
       call h5pcreate_f(h5p_dataset_xfer_f, xferid, hdferr)
       call h5pset_dxpl_mpio_f(xferid, h5fd_mpio_collective_f, hdferr)
    endif
#endif

    ! Create the global file space for the data

    call h5screate_simple_f(ndims, dims_file, fspcid, hdferr)
    
    ! Create the local memory space for the data

    call h5screate_simple_f(ndims, dimsf, mspcid, hdferr)

    ! If we are only writing a subset of the local data,
    ! select the points that we will output

    if (present(mcoords)) then

       if (size(mcoords) == 0) then

          call h5sselect_none_f(mspcid, hdferr)

       else
          
          if (ndims > 1) then
             offset(1:ndims-1) = 0
             countf(1:ndims-1) = dims(1:ndims-1)
          endif
          
          offset(ndims) = mcoords(1) - 1
          countf(ndims) = 1
          iop           = H5S_SELECT_SET_F

          ! Select a contiguous group of points as one slab for efficiency

          do j = 2, size(mcoords)
             if (mcoords(j) /= mcoords(j-1) + 1) then
                call h5sselect_hyperslab_f(mspcid, iop, offset, countf, hdferr)
                offset(ndims) = mcoords(j) - 1
                countf(ndims) = 1
                iop           = H5S_SELECT_OR_F
             else
                countf(ndims) = countf(ndims) + 1
             endif
          enddo

          call h5sselect_hyperslab_f(mspcid, iop, offset, countf, hdferr)
          
       endif
             
    endif

    ! Create the dataspace for the data

    call h5screate_simple_f(ndims, dims_file, dspcid, hdferr)

    ! For parallel output, select the points in the output file that
    ! we will be writing to

    if (present(fcoords) .and. present(ifsize)) then

       if (size(fcoords) == 0) then

          call h5sselect_none_f(dspcid, hdferr)

       else

          if (ndims > 1) then
             offset(1:ndims-1) = 0
             countf(1:ndims-1) = dims(1:ndims-1)
          endif
          
          offset(ndims) = fcoords(1) - 1
          countf(ndims) = 1
          iop           = H5S_SELECT_SET_F

          ! Select a contiguous group of points as one slab for efficiency

          do j = 2, size(fcoords)
             if (fcoords(j) /= fcoords(j-1) + 1) then
                call h5sselect_hyperslab_f(dspcid, iop, offset, countf, hdferr)
                offset(ndims) = fcoords(j) - 1
                countf(ndims) = 1
                iop           = H5S_SELECT_OR_F
             else
                countf(ndims) = countf(ndims) + 1
             endif
          enddo
          
          call h5sselect_hyperslab_f(dspcid, iop, offset, countf, hdferr)

       endif
             
    endif

    return
  end subroutine fh5_prepare_write



  subroutine fh5_write_integer_array(buf_integer, dname, hdferr)
    implicit none

    integer,      intent(IN)  :: buf_integer(*)
    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr

    call h5dcreate_f(fileid, dname, H5T_NATIVE_INTEGER, fspcid, dsetid, hdferr, propid)

    call h5dwrite_f(dsetid, H5T_NATIVE_INTEGER, buf_integer, dimsf, hdferr, mspcid, dspcid, xferid)

    return
  end subroutine fh5_write_integer_array



  subroutine fh5_write_real_array(buf_real, dname, hdferr)
    implicit none
 
    real,         intent(IN)  :: buf_real(*)
    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr

    call h5dcreate_f(fileid, dname, H5T_NATIVE_REAL, fspcid, dsetid, hdferr, propid)
    
    call h5dwrite_f(dsetid, H5T_NATIVE_REAL, buf_real, dimsf, hdferr, mspcid, dspcid, xferid)

    return
  end subroutine fh5_write_real_array



  subroutine fh5_write_character_array(buf_character, dname, hdferr)
    implicit none

    character(*), intent(IN)  :: buf_character(*)
    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr

    call h5dcreate_f(fileid, dname, H5T_NATIVE_CHARACTER, fspcid, dsetid, hdferr, propid)

    call h5dwrite_f(dsetid, H5T_NATIVE_CHARACTER, buf_character, dimsf, hdferr, mspcid, dspcid, xferid)

    return
  end subroutine fh5_write_character_array



  subroutine fh5_write_real8_array(buf_real8, dname, hdferr)
    implicit none

    real(kind=8), intent(IN)  :: buf_real8(*)
    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr

    call h5dcreate_f(fileid, dname, H5T_NATIVE_DOUBLE, fspcid, dsetid, hdferr, propid)

    call h5dwrite_f(dsetid, H5T_NATIVE_DOUBLE, buf_real8, dimsf, hdferr, mspcid, dspcid, xferid)

    return
  end subroutine fh5_write_real8_array



  subroutine fh5_write_logical_array(buf_logical, dname, hdferr)
    implicit none

    logical,      intent(IN)  :: buf_logical(*)
    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr
    integer,      allocatable :: buf_integer(:)
    integer                   :: i, size

    call h5dcreate_f(fileid, dname, H5T_NATIVE_INTEGER, fspcid, dsetid, hdferr, propid)

    ! converting logical to integer
    size = 1
    do i = 1, ndimsf
       size = size * dimsf(i)
    enddo

    allocate (buf_integer(size))
    buf_integer = 0

    do i = 1, size
       if(buf_logical(i)) buf_integer(i) = -1
    enddo

    call h5dwrite_f(dsetid, H5T_NATIVE_INTEGER, buf_integer, dimsf, hdferr, mspcid, dspcid, xferid)

    deallocate(buf_integer)

    return
  end subroutine fh5_write_logical_array



  subroutine fh5_write_integer_scalar(buf_integer, dname, hdferr)
    implicit none

    integer,      intent(IN)  :: buf_integer
    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr

    call h5dcreate_f(fileid, dname, H5T_NATIVE_INTEGER, fspcid, dsetid, hdferr, propid)

    call h5dwrite_f(dsetid, H5T_NATIVE_INTEGER, buf_integer, dimsf, hdferr, mspcid, dspcid, xferid)

    return
  end subroutine fh5_write_integer_scalar



  subroutine fh5_write_real_scalar(buf_real, dname, hdferr)
    implicit none

    real,         intent(IN)  :: buf_real
    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr

    call h5dcreate_f(fileid, dname, H5T_NATIVE_REAL, fspcid, dsetid, hdferr, propid)

    call h5dwrite_f(dsetid, H5T_NATIVE_REAL, buf_real, dimsf, hdferr, mspcid, dspcid, xferid)

    return
  end subroutine fh5_write_real_scalar



  subroutine fh5_write_character_scalar(buf_character, dname, hdferr)
    implicit none

    character(*), intent(IN)  :: buf_character
    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr

    call h5dcreate_f(fileid, dname, H5T_NATIVE_CHARACTER, fspcid, dsetid, hdferr, propid)

    call h5dwrite_f(dsetid, H5T_NATIVE_CHARACTER, buf_character, dimsf, hdferr, mspcid, dspcid, xferid)

    return
  end subroutine fh5_write_character_scalar



  subroutine fh5_write_real8_scalar(buf_real8, dname, hdferr)
    implicit none

    real(kind=8), intent(IN)  :: buf_real8
    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr

    call h5dcreate_f(fileid, dname, H5T_NATIVE_DOUBLE, fspcid, dsetid, hdferr, propid)

    call h5dwrite_f(dsetid, H5T_NATIVE_DOUBLE, buf_real8, dimsf, hdferr, mspcid, dspcid, xferid)

    return
  end subroutine fh5_write_real8_scalar



  subroutine fh5_write_logical_scalar(buf_logical, dname, hdferr)
    implicit none

    logical,      intent(IN)  :: buf_logical
    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr
    integer                   :: buf_integer

    call h5dcreate_f(fileid, dname, H5T_NATIVE_INTEGER, fspcid, dsetid, hdferr, propid)

    buf_integer = 0
    if(buf_logical) buf_integer = -1

    call h5dwrite_f(dsetid, H5T_NATIVE_INTEGER, buf_integer, dimsf, hdferr, mspcid, dspcid, xferid)

    return
  end subroutine fh5_write_logical_scalar



  subroutine fh5_close_write(hdferr)
    implicit none

    integer, intent(OUT) :: hdferr

    call h5sclose_f(fspcid, hdferr)
    call h5sclose_f(dspcid, hdferr)
    call h5sclose_f(mspcid, hdferr)

    call h5dclose_f(dsetid, hdferr)
    call h5pclose_f(xferid, hdferr)
    call h5pclose_f(propid, hdferr)

    return
  end subroutine fh5_close_write



  subroutine fh5d_open(dname, hdferr)
    implicit none

    character(*), intent(IN)  :: dname
    integer,      intent(OUT) :: hdferr

    call h5dopen_f(fileid, dname, dsetid, hdferr)

    if(dsetid < 0) then
       hdferr = int(dsetid)
       return
    end if

    call h5dget_space_f(dsetid, dspcid, hdferr)

    hdferr = int(dspcid) + hdferr

    return
  end subroutine fh5d_open



  subroutine fh5s_get_ndims(ndims)
    implicit none

    integer, intent(OUT) :: ndims

    integer :: hdferr

    call h5sget_simple_extent_ndims_f(dspcid, ndims, hdferr)

    return
  end subroutine fh5s_get_ndims



  subroutine fh5s_get_dims(dims)
    implicit none

    integer, intent(OUT) :: dims(:)
    integer(HSIZE_T)     :: maxdimsc(7), dimsc(7)
    integer              :: ndims, i
    integer              :: hdferr

    call h5sget_simple_extent_ndims_f(dspcid, ndims, hdferr)

    call h5sget_simple_extent_dims_f(dspcid, dimsc, maxdimsc, hdferr)

    do i = 1, ndims
       dims(i) = int(dimsc(i))
    end do

    return
  end subroutine fh5s_get_dims



  subroutine fh5d_close(hdferr)
    implicit none

    integer, intent(OUT) :: hdferr

    call h5sclose_f(dspcid, hdferr)
    call h5dclose_f(dsetid, hdferr)

    return
  end subroutine fh5d_close



  subroutine fh5d_read_integer_array(buf_integer, hdferr)
    implicit none

    integer, intent(INOUT) :: buf_integer(*)
    integer, intent(OUT)   :: hdferr

    call h5dread_f(dsetid, H5T_NATIVE_INTEGER, buf_integer, dimsf, &
         hdferr, file_space_id=dspcid, mem_space_id=mspcid)

    return
  end subroutine fh5d_read_integer_array



  subroutine fh5d_read_real_array(buf_real, hdferr)
    implicit none

    real,    intent(INOUT) :: buf_real(*)
    integer, intent(OUT)   :: hdferr

    call h5dread_f(dsetid, H5T_NATIVE_REAL, buf_real, dimsf, &
         hdferr, file_space_id=dspcid, mem_space_id=mspcid)

    return
  end subroutine fh5d_read_real_array



  subroutine fh5d_read_character_array(buf_character, hdferr)
    implicit none

    character, intent(INOUT) :: buf_character(*)
    integer,   intent(OUT)   :: hdferr

    call h5dread_f(dsetid, H5T_NATIVE_CHARACTER, buf_character, dimsf, &
         hdferr, file_space_id=dspcid, mem_space_id=mspcid)

    return
  end subroutine fh5d_read_character_array



  subroutine fh5d_read_real8_array(buf_real8, hdferr)
    implicit none

    real(KIND=8), intent(INOUT) :: buf_real8(*)
    integer,      intent(OUT)   :: hdferr

    call h5dread_f(dsetid, H5T_NATIVE_DOUBLE, buf_real8, dimsf, &
         hdferr, file_space_id=dspcid, mem_space_id=mspcid)

    return
  end subroutine fh5d_read_real8_array



  subroutine fh5d_read_logical_array(buf_logical, hdferr)
    implicit none

    logical, intent(INOUT) :: buf_logical(*)
    integer, intent(OUT)   :: hdferr
    integer, allocatable   :: buf_integer(:)
    integer                :: i, size

    hdferr = 1

    size = 1
    do i = 1, ndimsf
       size = size * dimsf(i)
    enddo
    allocate (buf_integer(size))

    call h5dread_f(dsetid, H5T_NATIVE_INTEGER, buf_integer, dimsf, &
         hdferr, file_space_id=dspcid, mem_space_id=mspcid)

    ! converting integer to logical
    do i = 1, size
       if(buf_integer(i) == -1) then
          buf_logical(i) = .true.
       else
          buf_logical(i) = .false.
       endif
    enddo

    deallocate(buf_integer)

    return
  end subroutine fh5d_read_logical_array



  subroutine fh5d_read_integer_scalar(buf_integer, hdferr)
    implicit none

    integer, intent(INOUT) :: buf_integer
    integer, intent(OUT)   :: hdferr

    call h5dread_f(dsetid, H5T_NATIVE_INTEGER, buf_integer, dimsf, hdferr)

    return
  end subroutine fh5d_read_integer_scalar



  subroutine fh5d_read_real_scalar(buf_real, hdferr)
    implicit none

    real,    intent(INOUT) :: buf_real
    integer, intent(OUT)   :: hdferr

    call h5dread_f(dsetid, H5T_NATIVE_REAL, buf_real, dimsf, hdferr)

    return
  end subroutine fh5d_read_real_scalar



  subroutine fh5d_read_character_scalar(buf_character, hdferr)
    implicit none

    character, intent(INOUT) :: buf_character
    integer,   intent(OUT)   :: hdferr

    call h5dread_f(dsetid, H5T_NATIVE_CHARACTER, buf_character, dimsf, hdferr)

    return
  end subroutine fh5d_read_character_scalar



  subroutine fh5d_read_real8_scalar(buf_real8, hdferr)
    implicit none

    real(KIND=8), intent(INOUT) :: buf_real8
    integer, intent(OUT)        :: hdferr

    call h5dread_f(dsetid, H5T_NATIVE_DOUBLE, buf_real8, dimsf, hdferr)

    return
  end subroutine fh5d_read_real8_scalar



  subroutine fh5d_read_logical_scalar(buf_logical, hdferr)
    implicit none

    logical, intent(INOUT) :: buf_logical
    integer, intent(OUT)   :: hdferr
    integer                :: buf_integer

    hdferr = 1

    call h5dread_f(dsetid, H5T_NATIVE_INTEGER, buf_integer, dimsf, hdferr)

    buf_logical = .false.
    if(buf_integer == -1) buf_logical = .true.

    return
  end subroutine fh5d_read_logical_scalar



  subroutine fh5_prepare_read(dname, ndims, dims, hdferr, coords, &
                              start, counts)
    implicit none

    character(*),      intent(IN)  :: dname
    integer,           intent(IN)  :: ndims
    integer,           intent(IN)  :: dims(:)
    integer,           intent(OUT) :: hdferr
    integer, optional, intent(IN)  :: coords(:)
    integer, optional, intent(IN)  :: start(:)
    integer, optional, intent(IN)  :: counts(:)
    integer                        :: i, j, k, ndims_file, iop
    integer(HSIZE_T)               :: dims_file(7), maxdims_file(7)
    integer(HSIZE_T)               :: offset(ndims), countf(ndims)

    dimsf  = 1
    ndimsf = ndims

    do i = 1, ndims
       dimsf(i) = dims(i)
    enddo

    ! opening the dataset from file

    call h5dopen_f(fileid, dname, dsetid, hdferr)
    if(dsetid < 0) then
       hdferr = int(dsetid)
       return
    end if

    ! getting the dataspace of the dataset (dimension, size, element type etc)

    call h5dget_space_f(dsetid, dspcid, hdferr)
    if(dspcid < 0) then
       hdferr = int(dspcid)
       return
    end if

    ! prepare the memspace to receive the data

    call h5screate_simple_f(ndims, dimsf, mspcid, hdferr)

    ! If coords is present as an argument, use it to select the points
    ! in the file "dataspace" that we want to read

    if (present(coords)) then

       ! report an error if coords not dimensioned correctly

       if (size(coords) /= dims(ndims)) then
          write(*,*) "Error in fh5_prepare_read:"
          stop       "Invalid number of points selected."
       endif

       ! check the dataspace dimensions in the file

       call h5sget_simple_extent_ndims_f(dspcid, ndims_file, hdferr)
       call h5sget_simple_extent_dims_f(dspcid, dims_file, maxdims_file, hdferr)

       if (size(coords) == 0) then

          call h5sselect_none_f(dspcid, hdferr)

       else if (dims(ndims) < dims_file(ndims)) then

          ! If the number of points we want is less than that in the file,
          ! select the points we want (else we read the entire dataspace)

          if (ndims > 1) then
             offset(1:ndims-1) = 0
             countf(1:ndims-1) = dims(1:ndims-1)
          endif
          
          offset(ndims) = coords(1) - 1
          countf(ndims) = 1
          iop           = H5S_SELECT_SET_F

          ! Select a contiguous group of points as one slab for efficiency

          do j = 2, size(coords)
             if (coords(j) /= coords(j-1) + 1) then
                call h5sselect_hyperslab_f(dspcid, iop, offset, countf, hdferr)
                offset(ndims) = coords(j) - 1
                countf(ndims) = 1
                iop           = H5S_SELECT_OR_F
             else
                countf(ndims) = countf(ndims) + 1
             endif
          enddo

          call h5sselect_hyperslab_f(dspcid, iop, offset, countf, hdferr)

       endif

    else if (present(start) .and. present(counts)) then

       ! If start and count are present, select the slab to read
       ! based on the start point (offset in file) and 
       ! the count (number to read in each dimension)
          
       if (size(start) >= ndims .and. size(counts) >= ndims) then
          
          offset(1:ndims) = start (1:ndims) - 1
          countf(1:ndims) = counts(1:ndims)

          if ( all(offset >= 0) .and. all(countf >= 1) ) then
             call h5sselect_hyperslab_f(dspcid, H5S_SELECT_SET_F, offset, countf, hdferr)
          else
             write(*,*) "Error in fh5_prepare_read:"
             write(*,*) "Invalid start and count values"
          endif

       else
          write(*,*) "Error in fh5_prepare_read:"
          write(*,*) "Invalid start and count dimensions"
       endif

    endif

  end subroutine fh5_prepare_read



  subroutine fh5_close_read(hdferr)
    implicit none

    integer, intent(OUT) :: hdferr

    call h5sclose_f(mspcid, hdferr)
    call h5sclose_f(dspcid, hdferr)
    call h5dclose_f(dsetid, hdferr)

  end subroutine fh5_close_read



end module hdf5_f2f
